

/* Module: Wavelets.
	Please note that this module is highly experimental.

	Credits: The core functions of the DWT, particularly the 
	subband filters, are leaning heavily on Numerical Recpies in C,
	second edition, chapter 13.

*/

/*		MODULE FOR CRYSTAL IMAGE 0.85 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */




#include <gtk/gtk.h>
#include <gdk/gdk.h>

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>



#include "../src/defs.h"
#include "../src/dlgbox.h"
#include "../src/functions.h"
#include "../src/findclust.h"
#include "../src/interface.h"
#include "../src/graphdsp.h"
#include "../src/dso.h"
#include "../src/macro.h"
#include "../src/gtkhelpers.h"
#include "../src/special1.h"
#include "../src/imgbuf.h"
#include "../src/application.h"
#include "../src/secondwnd.h"
#include "../src/tiffstuff.h"
#include "../src/roi.h"
#include "../src/viz.h"



typedef struct 
{
    int n_param;		/* Number of filter parameters */
    int hk_offs;		/* Parameter offset for scale filter centering, usually -n_param/2 */
    int gk_offs;		/* Parameter offset for wavelet filter centering, usually -n_param/2 */
    float *hk;		/* Array of scale filter parameters */
    float *gk;		/* Array of wavelet filter parameters */
} 
wavefilt;


typedef enum { W_NULL, W_HAAR, W_DAUB4, W_DAUB8, W_DAUB12, W_DAUB20, W_COIF6, W_COIF12 } WAVELET_TYPE;




/* Numerical Recipes median finder. TODO: Should go into the statistics
	module to help finding the n-th quantile etc.

	Returns the k-th smallest value in the float array data. The input array will be rearranged
	to have this value in location data[k], with all smaller elements moved to data[0..k-2] 
	(in arbitrary order) and all larger elements in data[k..n-1] (also in arbitrary order).
	NOTE: FUNCTION CHANGES ELEMENT ORDER OF INPUT ARRAY.
*/

#define WVSWAP(a,b) temp=(a);(a)=(b);(b)=temp;

float element_finder (float *data, int k, int n)
{
int i,ir,j,l,mid;
float a,temp;
unsigned char abort;
	
	l=0;
	ir=n-1;
	abort=FALSE;
	
	while (TRUE)					/* Exit this loop through the RETURN below */
	{
		if (ir <= l+1) 				/* Active partition contains 1 or 2 elements. */
		{
			if ((ir == l+1) && (data[ir] < data[l])) 		/* Case of 2 elements. */
			{
				WVSWAP (data[l], data[ir])
			}
			return data[k];
		} 
		else 
		{
			mid=(l+ir) >> 1; 						/* Choose median of left, center, and right elements... */
			WVSWAP(data[mid],data[l+1]);			/* ... as partitioning element a. */
			
			if (data[l] > data[ir]) 				/* Also, rearrange so that data[l]  data[l+1], data[ir]  data[l+1] */
				WVSWAP(data[l],data[ir]);
			
			if (data[l+1] > data[ir])
				WVSWAP(data[l+1],data[ir]);
			
			if (data[l] > data[l+1]) 
				WVSWAP(data[l],data[l+1]);

			i=l+1;									/* Initialize pointers for partitioning. */
			j=ir;
			a=data[l+1];							/* Partitioning element. */
			abort=FALSE;
			while (!abort)								/* Begin innermost loop */
			{
				do 
					i++; 							/* Scan up to find element > a.*/
				while (data[i] < a);
				do 
					j--; 							/* Scan down to find element < a. */
				while (data[j] > a);
				if (j < i) 
					abort=TRUE;					/* Pointers crossed. Partitioning complete. */
				else
					WVSWAP(data[i],data[j]);
			}

			data[l+1]=data[j];						/* Insert partitioning element. */
			data[j]=a;
			if (j >= k) ir=j-1;						/* Keep active the partition that contains the kth element. */
			if (j <= k) l=i;
		}
	}
}


#undef WVSWAP





/*

  ***    *   *   ****    ****      *     *   *   ****            *****   *       *****
 *   *   *   *   *   *   *   *    * *    *   *   *   *           *       *         *
 *       *   *   *   *   *   *   *   *   **  *   *   *           *       *         *
  ***    *   *   ****    ****    *   *   * * *   *   *           ****    *         *
     *   *   *   *   *   *   *   *****   *  **   *   *           *       *         *
 *   *   *   *   *   *   *   *   *   *   *   *   *   *           *       *         *
  ***     ***    ****    ****    *   *   *   *   ****            *       *****     *

*/

/* The subband filters perform one step of DWT decomposision or synthesis.
	These subband filters operate on an 1D array of float data (length
	must be a power of two or - less strictly - an even number) and return
	the scaling filter coefficients in the lower half of the array and the 
	wavelet coefficients in the upper half of the array.
	The variable idwt determines whether we want a forward DWT (idwt==0)
	or an inverse DWT (idwt!=0).
*/



void dead_subband (float *data, unsigned long n, int idwt, wavefilt dummy)
{
	dp (0, "Dead subband filter. Should not happen.\n");
}




/*	Daubechies-4 one-stage subband coding filter
		This filter takes a one-dimensional array of data [0...n-1], length n MUST be a power of two
		and performs one stage of subband coding (idwt=0 -> forward transform).  The output is interleaved,
		let's say the input is l0, l1, l2... (for Lambda), and the output is the next level's l0, l1, ...
		from the scaling filter and g0, g1, g2... from the wavelet filter. Then, the output has the
		sequence l0, g0, l1, g1, l2, g2... - However, the formulation here (two lines in the FOR loop)
		provides the data shuffling as well, i.e. the output is l0, l1, l2, ... , g0, g1, g2, ...
		whereby the sequence g0, g1, g2... starts at nh=n/2.
		
		Conversely, the inverse DWT (variable idwt=1) requires the sequence l0, l1, l2, ... , g0, g1, g2, ...
		as input and joins them to form the previous stage's l0, l1, l2, l3, ...
		
		Note that the length of the data array may be longer than n - in that case, the values above n-1
		remain untouched. This is important once one subband stange has executed and we have an array of
		l0, l1, l2, ... , g0, g1, g2, ... (i.e. reshuffled) - then we divide n by two and call
		the subband coding function again.
*/



void daub4_subband (float *data, unsigned long n, int idwt, wavefilt dummy)
{

float *buf;
unsigned long nh,nh1,i,j;
float cdaub4[4];
float fsqrt2, sqrt3;

	if (n < 4) return;
	
	buf= (float *) calloc (n+1, sizeof (float));		/* temporary storage buffer */
	nh = n >> 1;
	nh1 = (nh)+1;

	/* Use the explicit equations for the Daub-4 filter coefficients to obtain full precision 
		even if we'd decide to go for double rather than float. Note that this is an inefficiency
		as the coefficients get recomputed  every time we use this subband filter. */
	
	fsqrt2 = 4.0*sqrt(2.0); sqrt3 = sqrt(3.0);
	cdaub4[0] = (1+sqrt3)/fsqrt2;
	cdaub4[1] = (3+sqrt3)/fsqrt2;
	cdaub4[2] = (3-sqrt3)/fsqrt2;
	cdaub4[3] = (1-sqrt3)/fsqrt2;

	if (!idwt) 			/* Forward subband coding filter (forward one-level DWT). */
	{
		i=0;
		for (j=0; j<n-3; j+=2)
		{
			buf[i]    = cdaub4[0]*data[j] + cdaub4[1]*data[j+1] + cdaub4[2]*data[j+2] + cdaub4[3]*data[j+3];
			buf[i+nh] = cdaub4[3]*data[j] - cdaub4[2]*data[j+1] + cdaub4[1]*data[j+2] - cdaub4[0]*data[j+3];
			i++;
		}
		
		buf[i]    = cdaub4[0]*data[n-2] + cdaub4[1]*data[n-1] + cdaub4[2]*data[0] + cdaub4[3]*data[1];
		buf[i+nh] = cdaub4[3]*data[n-2] - cdaub4[2]*data[n-1] + cdaub4[1]*data[0] - cdaub4[0]*data[1];

	} 
	else 			/* Inverse filter (inverse DWT). */
	{
		buf[0] = cdaub4[2]*data[nh-1] + cdaub4[1]*data[n-1] + cdaub4[0]*data[0] + cdaub4[3]*data[nh];
		buf[1] = cdaub4[3]*data[nh-1] - cdaub4[0]*data[n-1] + cdaub4[1]*data[0] - cdaub4[2]*data[nh];
		j=2;
		
		for (i=0; i<nh-1; i++) 
		{

			buf[j++] = cdaub4[2]*data[i] + cdaub4[1]*data[i+nh] + cdaub4[0]*data[i+1] + cdaub4[3]*data[i+nh1];
			buf[j++] = cdaub4[3]*data[i] - cdaub4[0]*data[i+nh] + cdaub4[1]*data[i+1] - cdaub4[2]*data[i+nh1];
		}
	}
	for (i=0; i<n; i++) data[i]=buf[i];
	free (buf);

}


/* This one may still have issues, particularly with the high end. Matrix wraparound? */

void haar_subband (float *data, unsigned long n, int idwt, wavefilt dummy)
{

float *buf;
unsigned long nh,i,j;
float chaar[4];

	if (n < 4) return;
	
	chaar[0] = 1/sqrt(2);
	chaar[1] = chaar[0];
	
	buf= (float *) calloc (n+1, sizeof (float));		/* temporary storage buffer */
	nh = n >> 1;
	


	if (!idwt) 			/* Forward subband coding filter (forward one-level DWT). */
	{
		i=0;
		for (j=0; j<n-1; j+=2)
		{
			buf[i]    = chaar[0]*data[j] + chaar[1]*data[j+1];
			buf[i+nh] = chaar[1]*data[j] - chaar[0]*data[j+1];
			i++;
		}
		
	} 
	else 			/* Inverse filter (inverse DWT). */
	{
		j=0;
		for (i=0; i<nh-1; i++) 
		{

			buf[j++] = chaar[0]*data[i] + chaar[1]*data[i+nh];
			buf[j++] = chaar[1]*data[i] - chaar[0]*data[i+nh];
		}
	}
	for (i=0; i<n; i++) data[i]=buf[i];
	free (buf);

}


/*
	Initializing routine for general_subband. Here, the wavefilt structure is initialized and
	filled with the wavelet and scale parameters.
	This function MUST be called before general_subband can be used.
	Note that the above standalone functions make use of specific parameter set lengths
	and are therefore faster than general_subband.
*/


wavefilt wavelet_prepare (WAVELET_TYPE wlt)
{
int k;
float sgn;
wavefilt wfilt;

	/* scale filter (lowpass) parameters h(k). */

static float c_haar[2] = 	{	0.7071067118655, 0.7071067118655 };					/* Haar */

static float c_daub4[4] =	{	0.4829629131445341,0.8365163037378079,				/* DAUB4, 2 vanishing moments */
								0.2241438680420134,-0.1294095225512604 };

static float c_daub8[8] = 	{	 0.23037781330890,  0.71484657055292, 				/* DAUB8, 4 vanishing moments */
								 0.63088076792986, -0.02798376941686,
								-0.18703481171909,  0.03084138183556,
								 0.03288301166689, -0.01059740178507 };

static float c_daub12[12]=	{	0.111540743350, 0.494623890398, 0.751133908021,		/* DAUB12, 6 vanishing moments */
								0.315250351709,-0.226264693965,-0.129766867567,
								0.097501605587, 0.027522865530,-0.031582039318,
								0.000553842201, 0.004777257511,-0.001077301085 };

static float c_daub20[20] =	{	0.026670057901, 0.188176800078, 0.527201188932,		/* DAUB20, 10 vanishing moments */
								0.688459039454, 0.281172343661,-0.249846424327,
								-0.195946274377, 0.127369340336, 0.093057364604,
								-0.071394147166,-0.029457536822, 0.033212674059,
								0.003606553567,-0.010733175483, 0.001395351747,
								0.001992405295,-0.000685856695,-0.000116466855,
								0.000093588670,-0.000013264203 };

static float c_coif6[6] = 	{	-0.07273261951285,  0.33789766245781,				/* COIF6, 2 vanishing moments */
								 0.85257202021226,  0.38486484686420,
								-0.07273261951285, -0.01565572813546 };

static float c_coif12[12] =	{	 0.016387336463, -0.041464936781,					/* COIF12, 4 vanishing moments */
								-0.067372554722,  0.386110066823,
								 0.812723635449,  0.417005184423,
								-0.076488599078, -0.059434418646,
								 0.023680171946,  0.005611434819,
								-0.001823208870, -0.000720549446 };

	wfilt.n_param=0;
	
	if (wlt==W_HAAR) 
	{
		wfilt.n_param=2;
		wfilt.hk=c_haar;
	}
	else if (wlt==W_DAUB4) 
	{
		wfilt.n_param=4;
		wfilt.hk=c_daub4;
	}
	else if (wlt==W_DAUB8) 
	{
		wfilt.n_param=8;
		wfilt.hk=c_daub8;
	}
	else if (wlt==W_DAUB12) 
	{
		wfilt.n_param=12;
		wfilt.hk=c_daub12;
	}
	else if (wlt==W_DAUB20) 
	{
		wfilt.n_param=20;
		wfilt.hk=c_daub20;
	}
	else if (wlt==W_COIF6) 
	{
		wfilt.n_param=6;
		wfilt.hk=c_coif6;
	}
	else if (wlt==W_COIF12) 
	{
		wfilt.n_param=12;
		wfilt.hk=c_coif12;
	}
	else 
		ShowMessage ("Error setting up wavelet parameters");

	wfilt.gk = (float*) calloc (wfilt.n_param, sizeof (float));

	/* Generate the wavelet filter parameters g(k) from the scale filter h(k) */
	
	sgn = -1.0;
	for (k=0; k<wfilt.n_param; k++) 
	{
		wfilt.gk[wfilt.n_param-1-k]=sgn*wfilt.hk[k];
		sgn = -sgn;
	}
	
	dp (2, "Setting up filter parameters: You have:\n");
	dp (2, "Lowpass:  ");
	for (k=0; k<wfilt.n_param; k++) dp (2, "%7.4f ",wfilt.hk[k]); dp (2,"\n");
	dp (2, "Highpass: ");
	for (k=0; k<wfilt.n_param; k++) dp (2, "%7.4f ",wfilt.gk[k]); dp (2,"\n");
	

/*	
	These values center the "support" of the wavelets at each level. Alternatively, the "peaks"
	of the wavelets can be approximately centered by the choices hk_offs=-2 and gk_offs=-n+2.
	Note that daub4 and this function with n=4 use different default centerings.
*/
	wfilt.hk_offs = wfilt.gk_offs =  0; //-(n >> 1);
	return wfilt;

}



/*
	Generalized one-stage subband filter: applies an arbitrary wavelet filter to data  (for idwt==0)
	or applies its transpose (for idwt==1). Used hierarchically by the actual DWT functions.
	The actual filter is determined by a preceding (a required) call to wavelet_prepare, which initializes
	the structure wfilt.
*/


void general_subband (float *data, unsigned long n, int idwt, wavefilt wfilt)
{
float *buf;
unsigned long i,j,k,bitmask,ni,nj,nh,nmod;

	if (n < 4) return;
	
	buf= (float *) calloc (n+1, sizeof (float));		/* temporary storage buffer */

	nmod=wfilt.n_param*n; 						/* A positive constant equal to zero mod n. */
	bitmask=n-1;								/* Mask of all bits, since n is a power of 2. */
	nh = n >> 1;								/* Start of highpass data */

	if (!idwt)					/* Forward subband coding filter (forward one-level DWT). */
	{
		j=0;
		for (i=0; i<n; i+=2) 					/* Coefficient loop */
		{
			ni = i+nmod+wfilt.hk_offs;				/* Index into data to be incremented and wrapped-around. */
			nj = i+nmod+wfilt.gk_offs;				/* essentially i + a large multiple of n + offset */

			for (k=0; k<wfilt.n_param; k++)		/* Convolution loop */
			{
				buf[j]    += wfilt.hk[k]*data[(ni+k) & bitmask];	/* Scale filter (lowpass) data */
				buf[j+nh] += wfilt.gk[k]*data[(nj+k) & bitmask];	/* Wavelet filter (highpass) data */
			}
			j++;
		}
	} 
	else 			/* Inverse filter (inverse DWT). */
	{
		j=0;
		for (i=0; i<n; i+=2) 
		{
			ni = i+nmod+wfilt.hk_offs;
			nj = i+nmod+wfilt.gk_offs;
			for (k=0; k<wfilt.n_param; k++) 
			{
				buf[(ni+k) & bitmask] += wfilt.hk[k]*data[j];
				buf[(nj+k) & bitmask] += wfilt.gk[k]*data[j+nh];
			}
			j++;
		}
	}
	
	for (i=0; i<n; i++) data[i]=buf[i];        /* Copy the results back from workspace.*/
	free (buf);
}



/*********************************

 ####    #   #   #####
 #   #   #   #     #
 #   #   #   #     #
 #   #   # # #     #
 #   #   # # #     #
 #   #   # # #     #
 ####     # #      #

*************************************/


/*
	Here are the pyramidal deconstruction/reconstruction algorithms
	for 1D data, n-dimensional data etc. */


/*
	One-dimensional discrete wavelet transform. This routine implements the pyramid algorithm,
	replacing data[0..n-1] by its wavelet transform (for idwt=0), or performing the inverse operation
	(for idwt=1). Note that n MUST be an integer power of 2. The routine wt_subband, whose
	actual name must be supplied in calling this routine, is the underlying wavelet filter. Examples
	of wt_subband are daub4 and (preceded by wavelet_prepare) general_subband.
*/

void dwt_1d (float *data, unsigned long n, int idwt, wavefilt wfilt, void (*wt_subband)(float [], unsigned long, int, wavefilt))
{
unsigned long i;


	if (n < 4) return;

	if (!idwt)						/* Forward Wavelet transform. */
	{
		for (i=n; i>=4; i>>=1) 
			(*wt_subband)(data, i, idwt, wfilt);		/* Start at largest hierarchy, and work towards smallest.  */
		
	} 
	else 							/* Inverse wavelet transform.*/
	{
		for (i=4; i<=n; i<<=1) 
			(*wt_subband)(data, i, idwt, wfilt);		/* Start at smallest hierarchy, and work towards largest. */
	}
}


/* The same as above, but with a limited number of levels */

void dwt_1dn (float *data, unsigned long n, int idwt, int levels, wavefilt wfilt, void (*wt_subband)(float [], unsigned long, int, wavefilt))
{
unsigned long i;
int l;


	if (n < 4) return;

	if (!idwt)						/* Forward Wavelet transform. */
	{
		i=n; l=levels;
		while ((i>=4) && (l>1))
		{
			(*wt_subband)(data, i, idwt, wfilt);		/* Start at largest hierarchy, and work towards smallest.  */
			i >>= 1;
			l--;
		}
		
	} 
	else 							/* Inverse wavelet transform.*/
	{
		i = n; l=levels;
		while (l>1) { i>>=1; l--; }
		l = levels;
		while ((i<=n) && (l>1))
		{
			(*wt_subband)(data, i, idwt, wfilt);		/* Start at smallest hierarchy, and work towards largest. */
			i <<= 1;
			l--;
		}
	}
}



/* Simplistc slice-by-slice 2D wavelet transform. This function
	transforms the current image and returns a float image. */


void dwt_2d (image_type *img, int idwt, int decomp_level, wavefilt wfilt, void (*wt_subband)(float *, unsigned long, int, wavefilt))
{
image_type result;
float *buf1, *buf2, *p;
int nx, ny;
int x,y,z,i, level;


	/* Expand to nearest power of two */
	
	nx = 1; while (nx<img->xmax) nx=nx<<1;
	ny = 1; while (ny<img->ymax) ny=ny<<1;
	if ((nx<4) || (ny<4))
	{
		ShowMessage ("Not enough data for DWT");
		return;
	}

	/* Get memory */
	
	if (allocate_image (&result, B_FLOAT,nx, ny, img->zmax) ==-1)
		fatal_memerr();
	p = result.data;
	buf1 = calloc (nx, sizeof(float));
	buf2 = calloc (ny, sizeof(float));

	if (idwt) 
		dp (1, "2D inverse wavelet transform, %d level(s)...\n", decomp_level);
	else
		dp (1, "2D wavelet transform, %d level(s)...\n", decomp_level);


	for (z=0; z<img->zmax; z++)
	{

		/* Full pyramidal transform in X direction */

		for (y=0; y<img->ymax; y++)
		{
			for (x=0; x<nx; x++) buf1[x]=0;
			for (x=0; x<img->xmax; x++) buf1[x] = readbuf_flt (img, x,y,z);
		
			if (!idwt)						/* Forward Wavelet transform. */
			{
				i = nx; level=decomp_level;
				while ((i>=4) && (level>0))
				{
					dp (3, "2D DWT in X: Subband filter, level %d, size %d\n",level,i);
					(*wt_subband) (buf1, i, idwt, wfilt);		/* Start at largest hierarchy, and work towards smallest.  */
					i = i>>1;							/* Divide subband active range by two */
					level--;
				}
				
			} 
			else 							/* Inverse wavelet transform.*/
			{
				i=nx; level = decomp_level;
				while (level>1) { i>>=1; level--; }
				level = decomp_level;
				while ((i<=nx) && (level>0))
				{
					dp (3, "2D IDWT in X: Subband filter, level %d, size %d\n",level,i);
					(*wt_subband) (buf1, i, idwt, wfilt);		/* Start at smallest hierarchy, and work towards largest. */
					i = i<<1;
					level--;
				}
			}
	
			for (x=0; x<nx; x++) p[x + nx*(y+ny*z)] = buf1[x];
		}
	
		/* Full pyramidal transform in Y direction */
	
		for (x=0; x<nx; x++)
		{
			for (y=0; y<ny; y++) buf2[y]=p[x + nx*(y+ny*z)];
		
			if (!idwt)						/* Forward Wavelet transform. */
			{
				i = ny; level=decomp_level;
				while ((i>=4) && (level>0))
				{
					dp (3, "2D DWT in Y: Subband filter, level %d, size %d\n",level,i);
					(*wt_subband) (buf2, i, idwt, wfilt);		/* Start at largest hierarchy, and work towards smallest.  */
					i = i>>1;							/* Divide subband active range by two */
					level--;
				}
				
			} 
			else 							/* Inverse wavelet transform.*/
			{
				i=ny; level = decomp_level;
				while (level>1) { i>>=1; level--; }
				level = decomp_level;
				while ((i<=ny) && (level>0))
				{
					dp (3, "2D IDWT in Y: Subband filter, level %d, size %d\n",level,i);
					(*wt_subband) (buf2, i, idwt, wfilt);		/* Start at smallest hierarchy, and work towards largest. */
					i = i<<1;
					level--;
				}
			}
	
			for (y=0; y<ny; y++) p[x + nx*(y+ny*z)] = buf2[y];
		}
	}

	make_minmax (&result);
	handback (img, &result);

}



/*

 *   *   *   *   *       *****            ***    *   *   ****     ***    *   *   *   *    ***    *   *    ***
 *   *   *   *   *         *             *   *   *   *   *   *     *     *   *   *  *      *     *   *   *   *
 *   *   *   *   *         *             *       *   *   *   *     *     **  *   * *       *     **  *   *
 * * *    * *    *         *              ***    *****   ****      *     * * *   **        *     * * *   *  **
 * * *    * *    *         *                 *   *   *   * *       *     *  **   * *       *     *  **   *   *
 * * *     *     *         *             *   *   *   *   *  *      *     *   *   *  *      *     *   *   *   *
  * *      *     *****     *              ***    *   *   *   *    ***    *   *   *   *    ***    *   *    ***
 
*/


/* Wavelet shrinkage for denoising.

	2D slice-by-slice wavelet shrinkage with different threshold finding algorithms.
	img must be a wavelet-transformed image with a decomposition given in decomp_level.
	For this reason, we assume the image type to be B_FLOAT (otherwise, we return doing nothing)
	This function performs wavelet shrinkage according to wshrinkfunc and returns
	the shrunken coefficients for subsequent reconstruction.
	The function is
		0,1: Use threshold proportional to data length (MinMaxThresh)
		2,3: Determine median coefficient and use it as sigma with T=sigma*sqrt(2 log N)
		4,5: Use GCV to determine T

		even numbers are for soft thresholding, odd numbers for hard thresholding
	
*/



float wv_thresh (float elem, float t, float hardsoft)
{
	if (!hardsoft)			/* ==0 -> soft thresholding */
	{
		if (fabs(elem) <= fabs (t))
			return 0.0;
		else
			return SIGN(elem)*(fabs(elem)-fabs(t));
	}
	else					/* hard thresholding */
	{
		if (fabs(elem) > fabs (t))
			return elem;
		else
			return 0.0;
	}
}


/* Threshold method - threshold finding algorithms.
 	Note that x1,y1 defines the square area to operate on (usually the entire wavelet coeff area)
	while x2,y2 define the excluded are of L/L coefficients. In both cases, the highest value
 	is INCLUDED, e.g. for (x=0; x<=x1...) */


#define C06745		1.48258			/* this is 1/0.6745 */

float wvlt_find_threshold (image_type *img, int x1, int y1, int x2, int y2, int z, int wshrinkfunc)
{
int x,y,n,i,cnt;
float *p;				/* Intermediate linear array to hold data for sorting */
float w,T,min,max, buf, buf2;
float gcv, gcvmin, gcvmint, sumxx;

		n = (x1+1)*(y1+1)-(x2+1)*(y2+1);		/* Number of elements */

		switch (wshrinkfunc)
		{
			case 0:							/* MinMaxThresh */
			case 1:
				w = log(n)/log(2);			/* Log2 of data length */
				T = 0.185*(w-1.0);			/* Threshold */
				if (w<5)
					return 0.0;				/* Too small, don't threshold */
				else
					return T;
				break;
			case 2:							/* Median-based estimator for sigma */
			case 3:
				p = calloc (n, sizeof (float));
				i=0; min=max=fabs(readbuf_flt (img,x1,y1,z));
				#ifdef WV_FULLAREA_SIGMA
					for (y=0; y<=y1; y++)
						for (x=0; x<=x1; x++)
						{
							if ((x>x2) || (y>y2))			/* Use _all_ gamma coefficients, i.e. everything without L/L */
							{
								p[i++] = buf = fabs(readbuf_flt (img,x,y,z));
								min = MIN(min,buf);
								max = MAX(max,buf);
							}
						}
				#else
					for (y=y1/2+1; y<=y1; y++)				/* H/H area of level 1 only */
						for (x=x1/2+1; x<=x1; x++)
						{
							p[i++] = buf = fabs(readbuf_flt (img,x,y,z));
							min = MIN(min,buf);
							max = MAX(max,buf);
						}
				#endif

				w = element_finder (p, n/2, n);		/* Note that this implies an odd n (which we don't have), 
														but with large n the difference is minimal */
				T = w*C06745;						/* Donohan & Johnstone threshold = estimated standard deviation */
				dp (2, "      Median element from %d-vector is %f (min %f, max %f)  -> threshold %f\n",n,w,min,max,T);
				free (p);
				w = log(n)/log(2);					/* Log2 of data length */
				return T*sqrt(w);					/* return sigma * sqrt (2 log n) */
				break;
			case 4:							/* GCV minimization */
			case 5:
				i=0; min=max=fabs(readbuf_flt (img,x1,y1,z));
				for (y=0; y<=y1; y++)
					for (x=0; x<=x1; x++)
					{
						if ((x>x2) || (y>y2))				/* We need to know the lower and upper threshold bounds */
						{
							buf = fabs(readbuf_flt (img,x,y,z));
							max = MAX(max,buf);
						}
					}

				gcvmin=99999; gcvmint=0;
				max /= 5;									/* Nothing happens in the range of large coefficients */
				for (w=0.0; w<=max; w+=max/200.0)			/* span 0 to max in 200 steps. Good enough? */
				{
					sumxx = 0.0; cnt=0;
					for (y=0; y<=y1; y++)
						for (x=0; x<=x1; x++)
						{
							if ((x>x2) || (y>y2))				/* Test-threshold all coefficients */
							{
								buf = readbuf_flt (img,x,y,z);
								buf2 = wv_thresh (buf, w, 0); 			/* Always MUST use soft thresholding */
								sumxx += (buf-buf2)*(buf-buf2);			/* Squared difference between thresholded and original */
								if (buf2==0.0) cnt++;					/* Number of thesholded coefficients */
							}
						}
					if (cnt>0)
					{
						gcv = n*sumxx/cnt/cnt;
						if (gcv < gcvmin) { gcvmin=gcv; gcvmint=w; }
						dp (2, "    GCV finder: Threshold %f, gcv=%f\n",w,gcv);
					}
				}
				dp (2, "    --> GCV finder: Minimum GCV %f at threshold %f\n",gcvmin, gcvmint);
				return gcvmint;


		}

	return 0;			/* Silence a compiler warning */
}



void wavelet_shrink (image_type *img, int decomp_level, int wshrinkfunc, float tmult)
{
int x,y,z, xm,ym, x2,y2;
int level;
float *p;
float T;			/* The threshold */

	if (img->imgtype!=B_FLOAT) return;
	p = img->data;

	for (z=0; z<img->zmax; z++)
	{
		xm=img->xmax; ym=img->ymax;		/* Full image size in top level */
		x2=xm; y2=ym;

		if (wshrinkfunc<4)				/* Global thresholds */
		{
		
			for (level=0; level<decomp_level; level++)
			{
				x2 >>=1;
				y2 >>=1;
			}
				
			T = wvlt_find_threshold (img, xm-1,ym-1,x2-1,y2-1, z, wshrinkfunc);
			T *= tmult;
			dp (2, "Wavelet shrinking (slice %d): Area I, found threshold %f\n",z,T);
			for (y=0; y<ym; y++)
				for (x=0; x<xm; x++)
				{
					if ((x>=x2) || (y>=y2))
						p[x + img->xmax*(y + img->ymax*z)] = wv_thresh (p[x + img->xmax*(y + img->ymax*z)], T, wshrinkfunc & 0x01);
				}
		}
		else						/* Level-by-level thresholds */
		{
			xm=img->xmax; ym=img->ymax;             /* Full image size in top level */

			for (level=0; level<decomp_level; level++)
			{
				/* For each level, go over the H/L, H/H, and L/H areas separately */

				T = wvlt_find_threshold (img, xm-1,ym-1,xm/2-1,ym/2-1, z, wshrinkfunc);
				T *= tmult;
				dp (2, "Wavelet shrinking (slice %d, level %d): Found threshold %f\n",z,level,T);
				for (y=0; y<ym/2; y++)
					for (x=xm/2; x<xm; x++)
				{
					p[x + img->xmax*(y + img->ymax*z)] = wv_thresh (p[x + img->xmax*(y + img->ymax*z)], T, wshrinkfunc & 0x01);
				}

				for (y=ym/2; y<ym; y++)
					for (x=xm/2; x<xm; x++)
				{
					p[x + img->xmax*(y + img->ymax*z)] = wv_thresh (p[x + img->xmax*(y + img->ymax*z)], T, wshrinkfunc & 0x01);
				}

				for (y=ym/2; y<ym; y++)
					for (x=0; x<xm/2; x++)
				{
					p[x + img->xmax*(y + img->ymax*z)] = wv_thresh (p[x + img->xmax*(y + img->ymax*z)], T, wshrinkfunc & 0x01);
				}

				xm /=2; ym/=2;                          /* Shrink area by 2 for next level */

			}
		}


	}

}



/*
 *****   *****    ***    *****    ***
   *     *       *   *     *     *   *
   *     *       *         *     *
   *     ****     ***      *      ***
   *     *           *     *         *
   *     *       *   *     *     *   *
   *     *****    ***      *      ***

*/

/* A test function. Generate a test signal and decompose. Display
   transform and reconstruct, display again.
*/

void test_wavelet (wavefilt wfilt, void (*wt_subband)(float *, unsigned long, int, wavefilt) )
{
float xdata[256], ydata[256*3];
int i,N;
	N=256;

	for (i=0; i<N; i++)			/* Create some sort of signal */
	{
		xdata[i]=i;
		ydata[i]=0; ydata[i+N]=0; ydata[i+N+N]=0;
		if ((i>40) && (i<80)) ydata[i]=10+gauss_rnd()*2;		/* A noisy signal */
		if ((i>110) && (i<160)) ydata[i] = (i-110.0)/8.0;		/* A ramp */
		if (i>230) ydata[i]=5;									/* A step */
	}

	for (i=0; i<N; i++) ydata[i+N+N]=ydata[i];		/* Copy up to keep original signal */

	for(i=N; i>=64; i >>= 1)
		(*wt_subband)(ydata,i,0, wfilt);					/* two-level DWT decomposition */

	for (i=0; i<N; i++) ydata[i+N]=ydata[i];		/* Copy up to keep DWT */

	for(i=64; i<=N; i <<= 1)
		(*wt_subband)(ydata,i,1, wfilt);					/* Two-level IDWT */
		
	multi_graphme (xdata, ydata, N, 3);				/* Red: Reconstructed. Green: DWT, Blue: Original */


}



/*+**********************************+******************+

  ####    #    #   #*#
 #    #   #    #    #
 #        #    #    #
 #   ##   #    #    #
 #    #   #    #    #
 #    #   #    #    #
  ####     ####    #*#

*/


void (*subband)(float *, unsigned long, int, wavefilt);			/* Pointer to the selected subband filter function */
wavefilt wfilt;	
GtkWidget *wvlt_funcmenu;
GtkWidget *wvlt_shrinkmenu;
GtkWidget *wvlt_typemenu;
GtkWidget *wvlt_label4, *wvlt_label1, *wvlt_label5;
GtkWidget *wvlt_tmult;

int wfunction, woperation, wshrinkfunc, idwt;
GtkAdjustment *wvlt_decomp_adj, *wvlt_tmult_adj;




WAVELET_TYPE set_subband (int wavelet)
{
WAVELET_TYPE wtype;
	
	wtype = W_NULL;
	switch (wavelet)
	{
		case 0:						/* Haar */
				wtype=W_HAAR;
				subband = &general_subband;
				break;
		case 1:						/* DAUB4 */
				wtype=W_DAUB4;
				subband = &general_subband;
				break;
		case 2:						/* DAUB8 */
				wtype=W_DAUB8;
				subband = &general_subband;
				break;
		case 3:						/* DAUB12 */
				wtype=W_DAUB12;
				subband = &general_subband;
				break;
		case 4:						/* DAUB20 */
				wtype=W_DAUB20;
				subband = &general_subband;
				break;
		case 5:						/* Coiflet-6 */
				wtype=W_COIF6;
				subband = &general_subband;
				break;
		case 6:						/* Coiflet-12 */
				wtype=W_COIF12;
				subband = &general_subband;
				break;
		case 7:						/* Standalone  Haar */
				subband = &haar_subband;
				return W_NULL;
				break;
		case 8:						/* Standalone DAUB4 */
				subband = &daub4_subband;
				return W_NULL;
				break;
		
		default:
				subband = &dead_subband;
				return 0;
	}
	return wtype;
}


void on_wvlt_ok_clicked  (GtkButton *button, gpointer user_data)
{
int decomp_level;
WAVELET_TYPE wtype;
float tmult;

	undo_nexus (&mainimg);

	decomp_level = wvlt_decomp_adj->value;
	tmult = wvlt_tmult_adj->value;

	if ((woperation==0) || (woperation==1))						/* DWT and IDWT */
	{
		wtype = set_subband (wfunction);
		if (wtype!=W_NULL) wfilt = wavelet_prepare (wtype);
		dwt_2d (&mainimg, idwt, decomp_level, wfilt, *subband);
	}
	else if (woperation==2)										/* Wavelet shrinkage */
	{
		wavelet_shrink (&mainimg, decomp_level, wshrinkfunc, tmult);
	}

	make_imlib_buffer_from_main_image (cur_slice);

	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void wvlt_set_func_menu (GtkWidget *widget, gpointer user_data)
{
int func;

	func = *(int*) user_data;
	switch (func)
	{
		case 0:								/* Forward DWT */
				gtk_widget_hide (wvlt_label4); gtk_widget_hide (wvlt_shrinkmenu);
				gtk_widget_show (wvlt_label1); gtk_widget_show (wvlt_typemenu);
				gtk_widget_hide (wvlt_label5); gtk_widget_hide (wvlt_tmult);
				idwt=0; dp (2, "Changed IDWT flag to %d\n", idwt);
				break;
		case 1:								/* Inverse DWT */
				gtk_widget_hide (wvlt_label4); gtk_widget_hide (wvlt_shrinkmenu);
				gtk_widget_show (wvlt_label1); gtk_widget_show (wvlt_typemenu);
				gtk_widget_hide (wvlt_label5); gtk_widget_hide (wvlt_tmult);
				idwt=1; dp (2, "Changed IDWT flag to %d\n", idwt);
				break;
		case 2:								/* Wavelet shrinkage */
				gtk_widget_show (wvlt_label4); gtk_widget_show (wvlt_shrinkmenu);
				gtk_widget_hide (wvlt_label1); gtk_widget_hide (wvlt_typemenu);
				gtk_widget_show (wvlt_label5); gtk_widget_show (wvlt_tmult);
				break;
	}

	woperation = func;
}


void wvlt_set_shrink_menu (GtkWidget *widget, gpointer user_data)
{

	wshrinkfunc = *(int*) user_data;
	dp (2, "Changed shrinking threshold function to %d\n", wshrinkfunc);
	
}


void wvlt_set_type_menu (GtkWidget *widget, gpointer user_data)
{

	wfunction = *(int*) user_data;
	dp (2, "Changed wavelet to %d\n", wfunction);
	
}

void on_wvlt_cancel_clicked  (GtkButton *button, gpointer user_data)
{

	gtk_widget_destroy (GTK_WIDGET(user_data));
}


/* "Show" performs an inverse transform on a unit vector and thus generates the wavelet itself */

void on_wvlt_show_clicked  (GtkButton *button, gpointer user_data)
{
int M=2048;				/* lb 2048 = 11 levels */
float xdata[M], ydata[M+M];
int i;
WAVELET_TYPE wtype;

	wtype = set_subband (wfunction);
	if (wtype!=W_NULL) wfilt = wavelet_prepare (wtype);
	for (i=0; i<M; i++) {xdata[i]=i; ydata[i]=0; ydata[M+i]=0; }
	ydata[M+8] = 10;		/* Generates wavelet at 2 levels fewer than maximum - green */
	ydata[0] = 10; 			/* Generates scaling fn at 2 levels fewer than maximum - red */
	dwt_1dn (ydata, M, 1, 8, wfilt, (*subband));
	dwt_1dn (&ydata[M], M, 1, 8, wfilt, (*subband));
	multi_graphme (xdata, ydata, M, 2);

}


/* "Test" performs a full DWT/IDWT on a fixed 1D test signal */

void on_wvlt_test_clicked  (GtkButton *button, gpointer user_data)
{
WAVELET_TYPE wtype;

	wtype = set_subband (wfunction);
	if (wtype!=W_NULL) wfilt = wavelet_prepare (wtype);
	test_wavelet (wfilt, subband);

}




void create_wvlt_window (image_type *img)
{
  GtkWidget *wvlt_window;
  GtkWidget *vbox1, *hbox1, *wvlt_table;
  GtkWidget *wvlt_label2;
  GtkWidget *wvlt_label3;
  GtkWidget *hsep1;
  GtkWidget *wvlt_okbutton;
  GtkWidget *wvlt_showbutton;
  GtkWidget *wvlt_testbutton;
  GtkWidget *wvlt_cancelbutton;
  GtkWidget *wvlt_typemenu_menu;
  GtkWidget *wvlt_funcmenu_menu;
  GtkWidget *wvlt_shrinkmenu_menu;
  GtkWidget *glade_menuitem;
  GtkWidget *wvlt_decomp;

  float min,max;
  int y;

	imgminmax (img, &min, &max);

	wvlt_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (wvlt_window), "wvlt_window", wvlt_window);
	gtk_window_set_title (GTK_WINDOW (wvlt_window), "Wavelet Transform");

	vbox1 = vbox_in_window (wvlt_window,"vbox1"); 
	wvlt_table = MakeTable (wvlt_window, vbox1, "table1", 2,6);
	y=0;

	wvlt_label3 = MakeLabel (wvlt_window,wvlt_table,"Function","wvlt_label3",0,y);
	wvlt_funcmenu = Make_Optionmenu (wvlt_window, wvlt_table, "wvlt_funcmenu", 1,y);
	wvlt_funcmenu_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (wvlt_funcmenu_menu, GTK_SIGNAL_FUNC(wvlt_set_func_menu), "Forward DWT", (gpointer) &menuopt0);
	glade_menuitem = Option_menuitem (wvlt_funcmenu_menu, GTK_SIGNAL_FUNC(wvlt_set_func_menu), "Inverse DWT", (gpointer) &menuopt1);
	glade_menuitem = Option_menuitem (wvlt_funcmenu_menu, GTK_SIGNAL_FUNC(wvlt_set_func_menu), "Wavelet Shrinking", (gpointer) &menuopt2);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (wvlt_funcmenu), wvlt_funcmenu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (wvlt_funcmenu), 0); 
	gtk_container_set_border_width (GTK_CONTAINER (wvlt_funcmenu), 5);
	y++;
	idwt = 0; woperation=0;
	
	wvlt_label4 = MakeLabel (wvlt_window,wvlt_table,"Threshold","wvlt_label4",0,y);
	wvlt_shrinkmenu = Make_Optionmenu (wvlt_window, wvlt_table, "wvlt_shrinkmenu", 1,y);
	wvlt_shrinkmenu_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_set_shrink_menu), "MinMaxT (soft)", (gpointer) &menuopt0);
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_set_shrink_menu), "MinMaxT (hard)", (gpointer) &menuopt1);
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_set_shrink_menu), "Median coefficient (soft)", (gpointer) &menuopt2);
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_set_shrink_menu), "Median coefficient (hard)", (gpointer) &menuopt3);
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_set_shrink_menu), "GCVT (soft)", (gpointer) &menuopt4);
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_set_shrink_menu), "GCVT (hard)", (gpointer) &menuopt5);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (wvlt_shrinkmenu), wvlt_shrinkmenu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (wvlt_shrinkmenu), 0); idwt = 0;
	gtk_container_set_border_width (GTK_CONTAINER (wvlt_shrinkmenu), 5);
	gtk_widget_hide (wvlt_label4); gtk_widget_hide (wvlt_shrinkmenu);
	y++;

	wvlt_label1 = MakeLabel (wvlt_window,wvlt_table,"Wavelet type","wvlt_label1",0,y);
	wvlt_typemenu = Make_Optionmenu (wvlt_window, wvlt_table, "wvlt_typemenu", 1,y);
	wvlt_typemenu_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_set_type_menu), "Haar", (gpointer) &menuopt0);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_set_type_menu), "DAUB4", (gpointer) &menuopt1);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_set_type_menu), "DAUB8", (gpointer) &menuopt2);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_set_type_menu), "DAUB12", (gpointer) &menuopt3);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_set_type_menu), "DAUB20", (gpointer) &menuopt4);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_set_type_menu), "COIF6", (gpointer) &menuopt5);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_set_type_menu), "COIF12", (gpointer) &menuopt6);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_set_type_menu), "Haar (standalone)", (gpointer) &menuopt7);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_set_type_menu), "DAUB4 (standalone)", (gpointer) &menuopt8);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (wvlt_typemenu), wvlt_typemenu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (wvlt_typemenu), 0); wfunction=0;
	gtk_container_set_border_width (GTK_CONTAINER (wvlt_typemenu), 5);
	y++;

	wvlt_label2 = MakeLabel (wvlt_window,wvlt_table,"Decomposition level","wvlt_label2",0,y);
	wvlt_decomp_adj = GTK_ADJUSTMENT(gtk_adjustment_new (5, 1, 10, 1, 1, 0));
	wvlt_decomp = MakeHScale (wvlt_decomp_adj, wvlt_window, wvlt_table,"wvlt_decomp", 1,y);
	gtk_scale_set_digits (GTK_SCALE (wvlt_decomp), 0);
	y++;

	wvlt_label5 = MakeLabel (wvlt_window,wvlt_table,"Threshold Multiplier","wvlt_label5",0,y);
	wvlt_tmult_adj = GTK_ADJUSTMENT(gtk_adjustment_new (1, 0, 5, 0.1, 1, 0));
	wvlt_tmult = MakeHScale (wvlt_tmult_adj, wvlt_window, wvlt_table,"wvlt_tmult", 1,y);
	gtk_scale_set_digits (GTK_SCALE (wvlt_tmult), 1);
	gtk_widget_hide (wvlt_label5); gtk_widget_hide (wvlt_tmult);
	y++;

	
	hsep1 = MakeHsep (wvlt_window, wvlt_table, "hsep1", 0,y);
	hsep1 = MakeHsep (wvlt_window, wvlt_table, "hsep2", 1,y);
	y++;

	wvlt_showbutton = Button_in_table (wvlt_window, wvlt_table, "Show wavelet", "wvlt_showbutton",0,y, NULL);
	wvlt_testbutton = Button_in_table (wvlt_window, wvlt_table, "Test 1D DWT", "wvlt_testbutton",1,y, NULL);
	y++;

	hsep1 = MakeHsep (wvlt_window, vbox1, "hsep3", -1,-1);
	hbox1 = hbox_in_window (wvlt_window, vbox1, "wvlt_hbox");

	wvlt_okbutton = Button_in_table (wvlt_window, hbox1, "OK", "wvlt_okbutton",-1,-1, NULL);
	wvlt_cancelbutton = Button_in_table (wvlt_window, hbox1, "CANCEL", "wvlt_cancelbutton",-1,-1,NULL);

	gtk_signal_connect (GTK_OBJECT (wvlt_okbutton), "clicked", GTK_SIGNAL_FUNC (on_wvlt_ok_clicked), wvlt_window);
	gtk_signal_connect (GTK_OBJECT (wvlt_showbutton), "clicked", GTK_SIGNAL_FUNC (on_wvlt_show_clicked), wvlt_window);
	gtk_signal_connect (GTK_OBJECT (wvlt_testbutton), "clicked", GTK_SIGNAL_FUNC (on_wvlt_test_clicked), wvlt_window);
	gtk_signal_connect (GTK_OBJECT (wvlt_cancelbutton), "clicked", GTK_SIGNAL_FUNC (on_wvlt_cancel_clicked), wvlt_window);
	gtk_signal_connect (GTK_OBJECT (wvlt_window), "destroy", GTK_SIGNAL_FUNC (on_wvlt_cancel_clicked), wvlt_window);


	gtk_widget_show (wvlt_window); 

}


void wavelet_main ()
{

	create_wvlt_window (&mainimg);

}



/**************************************************************

	This is the macro interpreter function that hooks up
	to the macro interpreter. A pointer to this function gets
	passed to the macro registry function.


**************************************************************/

void dwt_macro_driver (image_type *img, WAVELET_TYPE wvlt, int idwt, int level)
{

	subband = &general_subband;			/* We always use the general_subband function in macros */
	wfilt = wavelet_prepare (wvlt);
	
	dwt_2d (img, idwt, level, wfilt, *subband);
	make_minmax (img);
}


/*
	DWT2D function
	Syntax: imgvar = DWT2D (imgexpr, "wavelet", idwt, level);
	Compute the discrete wavelet transform of imgexpr.
	"wavelet" is the wavelet type, one of HAAR, DAUB4, DAUB12, DAUB20...
	idwt decides between forward transform (idwt=0) and inverse transform (idwt=1)
	level is the decomposition level.
*/

variable_type mac_dwt (unsigned char* s, int* i)
{
variable_type v,r;
int j,k;
char wparm[20];
WAVELET_TYPE wvlt;
int idwt, level;


	/* The image stack */

	j = *i; k=0; v = nullvar;
	if (!isalpha(s[j])) macro_error (1,j);

	v = expr (s,&j, MT_IMG);		/* Get image to transform */
	
	/* The wavelet parameter */	

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_STRING);	strncpy (wparm, r.strg, 18); free (r.strg);
	wvlt = W_NULL;
	if (!strncasecmp (wparm, "haar",4)) wvlt = W_HAAR;
	else if (!strncasecmp (wparm, "daub4",5)) wvlt = W_DAUB4;
	else if (!strncasecmp (wparm, "daub8",5)) wvlt = W_DAUB8;
	else if (!strncasecmp (wparm, "daub12",6)) wvlt = W_DAUB12;
	else if (!strncasecmp (wparm, "daub20",6)) wvlt = W_DAUB20;
	else if (!strncasecmp (wparm, "coif6",5)) wvlt = W_COIF6;
	else if (!strncasecmp (wparm, "coif12",6)) wvlt = W_COIF12;
	else macro_error (15,j);	

	/* the integer arguments idwt and level */	

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); idwt = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_REAL); level = r.intvalue;

	while (s[j]!=')') j++;
	*i = j+1;

	/* Go execute. Prepare v to be returned as image */

	dwt_macro_driver (&v.imgptr, wvlt, idwt, level);
	v.type=MT_IMG;

	return v;
}



/**************************************************************

	This is the infrastructure we use to register with the 
	main program. The pointer "menufunction" points to the 
	main entry (the function to be called when the menu is 
	activated). Initmodule is the one function that the 
	main program calls (the name is fixed!) to initialize
	the module and to register the module with the menu.

***************************************************************/

void (*menufunction)();
variable_type (*macrofunction)(unsigned char* s, int* i);


void module_main()
{

	dp (1, "Wavelet module activated\n");
	if (file_ok) wavelet_main ();
}




void initmodule ()
{
	dp (1, "Wavelet module: Initializer called\n");
	
	/* Register this module with the main menu */
	
	menufunction = &module_main;
	dso_menu_register ("Wavelet", menufunction);
	
	/* Register the module with the macro interpreter */
	
	macrofunction = &mac_dwt;
	register_dso_macro ("DWT2D",4,MT_IMG,macrofunction);
	
}


/********************************************************************************/
