

/*
 *	processm: Handle some "Process" menu options
 */



/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <math.h>

#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>


#include "application.h"
#include "defs.h"
#include "dlgbox.h"
#include "functions.h"
#include "graphdsp.h"
#include "gtkhelpers.h"
#include "imgbuf.h"
#include "interface.h"
#include "miscop.h"
#include "processm.h"
#include "roi.h"
#include "secondwnd.h"


/*******************************************************

	Invert image. Well, just that.
	
********************************************************/


void invert_image (image_type* img)
{
long idx, numelms;

	numelms = img->xmax * img->ymax * img->zmax;

	switch (img->imgtype)
	{
		case B_CHAR:
		case B_RGB:
		{
			unsigned char* p;
			p = img->data;
			if (img->imgtype==B_RGB) numelms*=3;
			for (idx=0; idx<numelms; idx++)
			{
				p[idx] = img->maxval.imax - p[idx];
			}
			break;
		}
		case B_SHORT:
		{
			short* p;
			p = img->data;
			for (idx=0; idx<numelms; idx++)
			{
				p[idx] = img->maxval.imax - p[idx];
			}
			break;
		}
		case B_FLOAT:
		{
			float* p;
			p = img->data;
			for (idx=0; idx<numelms; idx++)
			{
				p[idx] = img->maxval.fmax - p[idx];
			}
			break;
		}
	}
	img->uid = next_uid++;
}



/*************************************************************

	Trim_Data does some adjustments to the image
	(executed in succession in this order):

	1. add / subtract offset
	2. shift the bits right or left
	3. AND them with a bitmask
	4. Clamp min and max values
	
	There is an unrelated (or somewhat related) function here,
	cycle_image to cyclically shift an image in x or y
	direction ("SpinCycle") - only allowed in 2D!
	
**************************************************************/



void cycle_image (image_type *img, int xcycle, int ycycle, int slice)
{
int x,y,e,xs,ys,xm;
long slicelen, sliceidx, cycleidx;
image_type tempimg;
unsigned char *p1, *p2;

	if ((xcycle==0) && (ycycle==0)) return;
	while (xcycle<0) xcycle+=img->xmax;
	while (ycycle<0) ycycle+=img->ymax;

	if (allocate_image (&tempimg, img->imgtype, img->xmax, img->ymax, 1)==-1)
		fatal_memerr();
	p1 = img->data;
	p2 = tempimg.data;

	e = DATALEN (img->imgtype);
	slicelen = img->xmax*img->ymax*e;		/* In bytes */
	sliceidx = slicelen * slice;			/* index to first byte */
	xcycle *= e;
	xm = img->xmax*e;
	
	if (xcycle)
	{
		memcpy (tempimg.data, &p1[sliceidx], slicelen);		/* Dup current slice to create source */
		for (y=0; y<img->ymax; y++)				/* x-cycle line by line */
			for (x=0; x<xm; x++)
			{
				xs = x-xcycle; while (xs<0) xs+=xm;
				cycleidx = xs+xm*y;
				p1[sliceidx+x+y*xm] = p2[cycleidx];
			}
	}

	if (ycycle)
	{
		memcpy (tempimg.data, &p1[sliceidx], slicelen);		/* Dup current slice to create source */
		for (y=0; y<img->ymax; y++)
		{
			ys = y-ycycle; while (ys<0) ys+=img->ymax;
			cycleidx = xm*ys;
			for (x=0; x<xm; x++)
				p1[sliceidx+x+y*xm] = p2[cycleidx+x];
		}
	}

	freebuf (&tempimg);
	img->uid = next_uid++;
}


void execute_trim_image (image_type* img, int offs, double fmin, double fmax, int shift, int mask)
{
long idx,i;
int imin,imax;
char minmaxflg;


	imin=floor(0.499+fmin); imax=floor(0.499+fmax);
	if ((fmin==0) && (fmax==0)) minmaxflg=1; else minmaxflg=0;
	idx = img->xmax * img->ymax * img->zmax;
	if (img->imgtype == B_RGB) idx *= sizeof (rgbtriplet);

	switch (img->imgtype)
	{
		case B_CHAR:
		case B_RGB:
		{
			unsigned char* p;
			unsigned char b;
			p = img->data;
			if (minmaxflg) imin=imax=*p;
			for (i=0; i<idx; i++)
			{
				b = p[i]-offs;
				if (shift>0) b = b >> shift;
				else if (shift<0) b = b << -shift;
				b = b & mask;
				if ( minmaxflg )
				{
					if (b<imin) imin=b;
					if (b>imax) imax=b;
				}
				p[i] = b;
			}
			break;
		}
		case B_SHORT:
		{
			short* p;
			short x;
			p = img->data;
			if (minmaxflg) imin=imax=*p;
			for (i=0; i<idx; i++)
			{
				x = p[i]-offs;
				if (shift>0) x = x >> shift;
				else if (shift<0) x = x << -shift;
				x = x & mask;
				if ( minmaxflg )
				{
					if (x<imin) imin=x;
					if (x>imax) imax=x;
				}
				p[i] = x;
			}
			break;
		}
		case B_FLOAT:
		{
			float* p;
			float f;
			p = img->data;
			if (minmaxflg) fmin=fmax=*p;
			for (i=0; i<idx; i++)
			{
				f = p[i]-offs;
				/* No shift & mask operations for float! */
				if ( minmaxflg )
				{
					if (f<fmin) fmin=f;
					if (f>fmax) fmax=f;
				}
				p[i] = f;
			}
			break;
		}
		
	}
		
	if (img->imgtype==B_FLOAT)
	{
		img->minval.fmin = fmin;
		img->maxval.fmax = fmax;
	}
	else
	{
		img->minval.imin=imin;
		img->maxval.imax = imax;
	}

	/* Then update the imlib buffer image */
	
	img->uid = next_uid++;

}









GtkWidget *trim_spinbutton1;
GtkWidget *trim_spinbutton2;
GtkWidget *trim_spinbutton3;
GtkWidget *trim_spinbutton4;
GtkWidget *trim_spinbutton5;
GtkWidget *trim_spinbutton6;
GtkWidget *trim_entry1;



void TrimOKClicked( GtkButton * button, gpointer data )
{

double fmin,fmax;
int shift;
int mask, offs;
int xcycle, ycycle,z;
char maskstr[100];

	undo_nexus (&mainimg);

	offs = gtk_spin_button_get_value_as_int ((GtkSpinButton*)trim_spinbutton4);
	fmin = gtk_spin_button_get_value_as_float ((GtkSpinButton*)trim_spinbutton1);
	fmax = gtk_spin_button_get_value_as_float ((GtkSpinButton*)trim_spinbutton2);
	shift= gtk_spin_button_get_value_as_int   ((GtkSpinButton*)trim_spinbutton3);
	xcycle= gtk_spin_button_get_value_as_int   ((GtkSpinButton*)trim_spinbutton5);
	ycycle= gtk_spin_button_get_value_as_int   ((GtkSpinButton*)trim_spinbutton6);
	strcpy (maskstr,gtk_entry_get_text (GTK_ENTRY(trim_entry1)));
	
	if (maskstr[0])
		sscanf (maskstr, "%x", &mask);
	else
		mask=0xffff;
	
	dp (1, "Trim Image: Min=%f, Max=%f, Offset=%d, Shift=%d, Mask=%04X\n",
				fmin,fmax,offs,shift,mask);
	
	execute_trim_image (&mainimg, offs, fmin,fmax,shift,mask);
	
	if (xcycle || ycycle)
	{
		dp (1, "Trim Image (2): Cycle in x: %d, cycle in y: %d\n",xcycle, ycycle);
		for (z=0; z<mainimg.zmax; z++)
			cycle_image (&mainimg, xcycle, ycycle, z);
	}
	
	make_imlib_buffer_from_main_image (cur_slice);
	redraw (1);

   	gtk_widget_destroy ( GTK_WIDGET ( data ) );
}


void TrimCancelClicked( GtkButton * button, gpointer data )
{

   gtk_widget_destroy ( GTK_WIDGET ( data ) );
}



void trim_data (image_type *img)
{
  GtkWidget *trimwnd, *vbox;
  GtkWidget *trim_table1;
  GtkObject *trim_spinbutton1_adj;
  GtkObject *trim_spinbutton2_adj;
  GtkObject *trim_spinbutton3_adj;
  GtkObject *trim_spinbutton4_adj;
  GtkObject *trim_spinbutton5_adj;
  GtkObject *trim_spinbutton6_adj;
  GtkWidget *trim_cancel;
  GtkWidget *trim_ok;
  GtkWidget *trim_label1;
  GtkWidget *trim_label2;
  GtkWidget *trim_label3;
  GtkWidget *trim_label4;
  GtkWidget *trim_label5;
  GtkWidget *trim_label6;
  GtkWidget *trim_label7;
  float min, max;
  
  imgminmax (img, &min, &max);

  trimwnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_object_set_data (GTK_OBJECT (trimwnd), "trimwnd", trimwnd);
  gtk_window_set_title (GTK_WINDOW (trimwnd), "Trim Image");
  gtk_window_set_modal (GTK_WINDOW (trimwnd), TRUE);

  vbox = vbox_in_window (trimwnd,"trim_vbox1");
  trim_table1 = MakeTable (trimwnd, vbox, "trim_table1",2,8);

  trim_label1 = MakeLabel (trimwnd, trim_table1, "Minimum", "trim_label1", 0,0);
  trim_spinbutton1_adj = gtk_adjustment_new (0, -32767, 32767, 1, 10, 10);
  trim_spinbutton1 = MakeSpinbutton (GTK_ADJUSTMENT(trim_spinbutton1_adj), trimwnd, trim_table1, "trim_spinbutton1", 1,0);

  trim_label2 = MakeLabel (trimwnd, trim_table1, "Maximum", "trim_label2", 0,1);
  trim_spinbutton2_adj = gtk_adjustment_new (0, -32767, 32767, 1, 10, 10);
  trim_spinbutton2 = MakeSpinbutton (GTK_ADJUSTMENT(trim_spinbutton2_adj), trimwnd, trim_table1, "trim_spinbutton2", 1,1);

  trim_label5 = MakeLabel (trimwnd, trim_table1, "Subtract offset", "trim_label5", 0,2);
  trim_spinbutton4_adj = gtk_adjustment_new (0, -4096, 4096, 1, 256, 256);
  trim_spinbutton4 = MakeSpinbutton (GTK_ADJUSTMENT(trim_spinbutton4_adj), trimwnd, trim_table1, "trim_spinbutton4", 1,2);

  trim_label3 = MakeLabel (trimwnd, trim_table1, "Shift right", "trim_label3", 0,3);
  trim_spinbutton3_adj = gtk_adjustment_new (0, -8, 8, 1, 4, 4);
  trim_spinbutton3 = MakeSpinbutton (GTK_ADJUSTMENT(trim_spinbutton3_adj), trimwnd, trim_table1, "trim_spinbutton3", 1,3);

  trim_label4 = MakeLabel (trimwnd, trim_table1, "Bitmask (hex)", "trim_label4", 0,4);
  trim_entry1 = Entry_in_table (trimwnd,trim_table1,NULL,"trim_entry1",1,4);

  trim_label6 = MakeLabel (trimwnd, trim_table1, "Cycle in X", "trim_label6", 0,5);
  trim_spinbutton5_adj = gtk_adjustment_new (0, 1-img->xmax, img->xmax-1, 1, 10, 10);
  trim_spinbutton5 = MakeSpinbutton (GTK_ADJUSTMENT(trim_spinbutton5_adj), trimwnd, trim_table1, "trim_spinbutton5", 1,5);

  trim_label7 = MakeLabel (trimwnd, trim_table1, "Cycle in Y", "trim_label7", 0,6);
  trim_spinbutton6_adj = gtk_adjustment_new (0, 1-img->ymax, img->ymax-1, 1, 10, 10);
  trim_spinbutton6 = MakeSpinbutton (GTK_ADJUSTMENT(trim_spinbutton6_adj), trimwnd, trim_table1, "trim_spinbutton6", 1,6);

  trim_cancel = Button_in_table (trimwnd,trim_table1,"Cancel","trim_cancel",1,7,NULL);
  trim_ok = Button_in_table (trimwnd,trim_table1,"OK","trim_ok",0,7,NULL);

  gtk_signal_connect (GTK_OBJECT (trim_ok), "clicked",
                      GTK_SIGNAL_FUNC (TrimOKClicked),trimwnd);
  gtk_signal_connect (GTK_OBJECT (trim_cancel), "clicked",
                      GTK_SIGNAL_FUNC (TrimCancelClicked),trimwnd);
  gtk_signal_connect (GTK_OBJECT (trimwnd), "destroy",
                      GTK_SIGNAL_FUNC (TrimCancelClicked),trimwnd);


  gtk_widget_show (trimwnd);
  
}


/**************************************************************

	Crop image
	
***************************************************************/


int do_crop (image_type* img, int x1, int y1, int x2, int y2)
{
image_type aux;
int y,z,ox,oy,l;


	if ((x1==x2) || (y2==y1)) return -1;
	if (x1>x2) { y=x1; x1=x2; x2=y; }
	if (y1>y2) { y=y1; y1=y2; y2=y; }

	/* If any of the coordinates x1 or y1 are negative,
	   we need to offset this into the dest image */
	
	ox = 0; if (x1<0) ox=-x1;
	oy = 0; if (y1<0) oy=-y1;
	if ((x2<0) || (y2<0)) return -2;
	l = x2-x1+1; if (l>img->xmax) l=img->xmax;	/* num elements per scanline */

	/* Retain type and z-depth, but have new x/y dimensions */
	
	if (allocate_image (&aux, img->imgtype, x2-x1+1, y2-y1+1, img->zmax)==-1)
		fatal_memerr();
		
	for (z=0; z<img->zmax; z++)
		for (y=y1+oy; y<=y2+oy; y++)
			if (y<img->ymax)
				copybuf (img, &aux, x1+ox,y,z, ox,y+oy-y1,z, l);
	
	/* Let the aux image take over */

	handback (img, &aux);	

	return 0;

}



int crop_leftmarg, crop_rightmarg, crop_topmarg, crop_bottommarg;


void on_crop_cancel_clicked ( GtkButton * button, gpointer data )
{

   gtk_widget_destroy ( GTK_WIDGET ( data ) );
}


void on_crop_ok_clicked ( GtkButton * button, gpointer data )
{

	undo_nexus (&mainimg);

	dp (1, "Cropping: Crop area is %d...%d by %d...%d\n",crop_leftmarg, crop_topmarg, crop_rightmarg, crop_bottommarg);
	do_crop (&mainimg, crop_leftmarg, crop_topmarg, crop_rightmarg, crop_bottommarg);
	dp (1, "Cropping: New image size is %li x %li x %li\n",mainimg.xmax,mainimg.ymax,mainimg.zmax);

	make_imlib_buffer_from_main_image (cur_slice);
	kill_roi();
	redraw (1);
	gtk_widget_destroy ( GTK_WIDGET ( data ) );
}

void on_crop_x1_adj_changed (GtkEditable *editable, gpointer user_data)
{
GtkAdjustment* adj;

	adj = user_data;
	crop_leftmarg = adj->value;
}

void on_crop_x2_adj_changed (GtkEditable *editable, gpointer user_data)
{
GtkAdjustment* adj;

	adj = user_data;
	crop_rightmarg = adj->value;
}

void on_crop_y1_adj_changed (GtkEditable *editable, gpointer user_data)
{
GtkAdjustment* adj;

	adj = user_data;
	crop_topmarg = adj->value;
}

void on_crop_y2_adj_changed (GtkEditable *editable, gpointer user_data)
{
GtkAdjustment* adj;

	adj = user_data;
	crop_bottommarg = adj->value;
}



void create_cropwnd (int x1, int y1, int x2, int y2)
{
  GtkWidget *cropwnd;
  GtkWidget *crop_vbox1;
  GtkWidget *crop_label1;
  GtkWidget *crop_table1;
  GtkWidget *crop_label2;
  GtkWidget *crop_label3;
  GtkWidget *crop_label4;
  GtkWidget *crop_label5;
  GtkWidget *crop_okbutton;
  GtkWidget *crop_cancelbutton;
  GtkObject *crop_x1_adj;
  GtkWidget *crop_x1;
  GtkObject *crop_y1_adj;
  GtkWidget *crop_y1;
  GtkObject *crop_x2_adj;
  GtkWidget *crop_x2;
  GtkObject *crop_y2_adj;
  GtkWidget *crop_y2;

  cropwnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_object_set_data (GTK_OBJECT (cropwnd), "cropwnd", cropwnd);
  gtk_window_set_title (GTK_WINDOW (cropwnd), "Crop image");
  gtk_window_set_modal (GTK_WINDOW (cropwnd), TRUE);

  crop_vbox1 = gtk_vbox_new (FALSE, 0);
  gtk_widget_ref (crop_vbox1);
  gtk_object_set_data_full (GTK_OBJECT (cropwnd), "crop_vbox1", crop_vbox1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (crop_vbox1);
  gtk_container_add (GTK_CONTAINER (cropwnd), crop_vbox1);

  crop_label1 = gtk_label_new ("Absolute coordinates to crop:");
  gtk_widget_ref (crop_label1);
  gtk_object_set_data_full (GTK_OBJECT (cropwnd), "crop_label1", crop_label1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (crop_label1);
  gtk_box_pack_start (GTK_BOX (crop_vbox1), crop_label1, FALSE, FALSE, 0);
  gtk_misc_set_padding (GTK_MISC (crop_label1), 4, 4);

  crop_table1 = gtk_table_new (5, 2, FALSE);
  gtk_widget_ref (crop_table1);
  gtk_object_set_data_full (GTK_OBJECT (cropwnd), "crop_table1", crop_table1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (crop_table1);
  gtk_box_pack_start (GTK_BOX (crop_vbox1), crop_table1, TRUE, TRUE, 0);

  crop_label2 = MakeLabel (cropwnd, crop_table1, "X1 (left)", "crop_label2", 0,0);
  crop_label3 = MakeLabel (cropwnd, crop_table1, "Y1 (top)", "crop_label3", 0,1);
  crop_label4 = MakeLabel (cropwnd, crop_table1, "X2 (right)", "crop_label4", 0,2);
  crop_label5 = MakeLabel (cropwnd, crop_table1, "Y2 (bottom)", "crop_label5", 0,3);

  crop_okbutton = Button_in_table (cropwnd, crop_table1, "OK (Crop)", "crop_okbutton", 0,4, NULL);
  crop_cancelbutton = Button_in_table (cropwnd, crop_table1, "Cancel", "crop_okbutton", 1,4, NULL);

  crop_x1_adj = gtk_adjustment_new (x1, -512, mainimg.xmax-1, 1, 10, 1);
  crop_x1 = MakeSpinbutton (GTK_ADJUSTMENT(crop_x1_adj), cropwnd, crop_table1, "crop_x1", 1,0);
  gtk_signal_connect (GTK_OBJECT (crop_x1), "value-changed", GTK_SIGNAL_FUNC (on_crop_x1_adj_changed),crop_x1_adj);
  crop_leftmarg = x1;

  crop_y1_adj = gtk_adjustment_new (y1, -512, mainimg.ymax-1, 1, 10, 1);
  crop_y1 = MakeSpinbutton (GTK_ADJUSTMENT(crop_y1_adj), cropwnd, crop_table1, "crop_y1", 1,1);
  gtk_signal_connect (GTK_OBJECT (crop_y1), "value-changed", GTK_SIGNAL_FUNC (on_crop_y1_adj_changed),crop_y1_adj);
  crop_topmarg = y1;

  crop_x2_adj = gtk_adjustment_new (x2, 0, mainimg.xmax+511, 1, 10, 1);
  crop_x2 = MakeSpinbutton (GTK_ADJUSTMENT(crop_x2_adj), cropwnd, crop_table1, "crop_x2", 1,2);
  gtk_signal_connect (GTK_OBJECT (crop_x2), "value-changed", GTK_SIGNAL_FUNC (on_crop_x2_adj_changed),crop_x2_adj);
  crop_rightmarg = x2;

  crop_y2_adj = gtk_adjustment_new (y2, 0, mainimg.ymax+511, 1, 10, 1);
  crop_y2 = MakeSpinbutton (GTK_ADJUSTMENT(crop_y2_adj), cropwnd, crop_table1, "crop_y2", 1,3);
  gtk_signal_connect (GTK_OBJECT (crop_y2), "value-changed", GTK_SIGNAL_FUNC (on_crop_y2_adj_changed),crop_y2_adj);
  crop_bottommarg = y2;


  gtk_signal_connect (GTK_OBJECT (crop_okbutton), "clicked", 
                      GTK_SIGNAL_FUNC (on_crop_ok_clicked), cropwnd);
  gtk_signal_connect (GTK_OBJECT (crop_cancelbutton), "clicked", 
                      GTK_SIGNAL_FUNC (on_crop_cancel_clicked), cropwnd);
  gtk_signal_connect (GTK_OBJECT (cropwnd), "destroy",
                      GTK_SIGNAL_FUNC (on_crop_cancel_clicked), cropwnd);


  gtk_widget_show (cropwnd);
  
}


void crop_main ()
{
int bx1,by1,bx2,by2;


	/* If we have a ROI, observe it */
	
	roi_bbox (&mainimg, &bx1,&by1,&bx2,&by2);
	create_cropwnd (bx1,by1,bx2,by2);

}



/****************************************************

	Histogram equalization
	
*****************************************************/


void histeq (image_type* img, int numbins, int oorflg)
{
int bx1,by1,bx2,by2;
int i,x,y,z,area,cumul,bin,garea,barea,c;
int bins[3*(numbins+1)];
float xlat[3*(numbins+1)];
float min, max, delta, t, buf;
float Sx,Sxx;
rgbtriplet rgb;
rgbtriplet *rgbptr;


	/* If we have a ROI, observe it */
	
	roi_bbox (img, &bx1,&by1,&bx2,&by2);

	imgminmax (img,&min,&max);
	rgbptr = img->data;

	for (i=0; i<=3*numbins; i++)
	{
		bins[i]=0; xlat[i]=0;
	}

	/* read the image */
	
	delta = (max-min)/numbins;	/* The width of one bin, starting with min */
	area=0; Sx=0; Sxx=0; garea=barea=0;
	for (z=0; z<img->zmax; z++)
		for (y=by1; y<=by2; y++)
			for (x=bx1; x<=bx2; x++)
			{
				if (in_roi (x,y))
				{
					buf = readbuf_flt (img,x,y,z);
					if ((!oorflg) || (fabs(buf)>1e-12))
					{
						if (img->imgtype == B_RGB)
						{
							rgb = rgbptr[x+img->xmax*(y+img->ymax*z)];
							bin = (int) ((rgb.red - min)/delta);
							if ((bin>=0) && (bin<=numbins)) 
							{
								bins[bin]++; area++;
							}
							bin = (int) ((rgb.green - min)/delta);
							if ((bin>=0) && (bin<=numbins)) 
							{
								bins[bin+numbins]++; garea++;
							}
							bin = (int) ((rgb.blue - min)/delta);
							if ((bin>=0) && (bin<=numbins)) 
							{
								bins[bin+2*numbins]++; barea++;
							}
							
						}
						else
						{
							bin = (int) ((buf - min)/delta);
							if ((bin>=0) && (bin<=numbins)) 
							{
								bins[bin]++;
								area++;
								Sx += buf; Sxx+= SQR(buf);		/* Just for fun - we don't use it in this function */
							}
						}
					}
				}
			}

	/* Compute the cumulative histogram which is also the translation table.
	   cumul/area ->1 when i->numbins */
	
	cumul=0;
	for (i=0; i<=numbins; i++)
	{
		cumul+=bins[i];
		xlat[i] = min+(max-min)*(float)cumul/(float)area;
		dp (3, "Bin %d value %f   P=%d, H=%f\n",i,i*delta+min,bins[i],xlat[i]);
	}
	
	/* Repeat for G and B if we have a RGB image */

	if (img->imgtype == B_RGB)
	{
		cumul=0;
		for (i=0; i<=numbins; i++)
		{
			cumul+=bins[i+numbins];
			xlat[i+numbins] = min+(max-min)*(float)cumul/(float)garea;
			dp (3, "Green Bin %d value %f   P=%d, H=%f\n",i,i*delta+min,bins[i+numbins],xlat[i+numbins]);
		}
		cumul=0;
		for (i=0; i<=numbins; i++)
		{
			cumul+=bins[i+2*numbins];
			xlat[i+2*numbins] = min+(max-min)*(float)cumul/(float)barea;
			dp (3, "Green Bin %d value %f   P=%d, H=%f\n",i,i*delta+min,bins[i+2*numbins],xlat[i+2*numbins]);
		}
	}
	

	/* Translate point by point, interpolate between bins */

	for (z=0; z<img->zmax; z++)
		for (y=by1; y<=by2; y++)
			for (x=bx1; x<=bx2; x++)
			{
				if (in_roi (x,y))
				{
					buf = readbuf_flt (img,x,y,z);
					if ((!oorflg) || (fabs(buf)>1e-12))
					{
						bin = (int) ((buf - min)/delta);
						t = (buf - (bin*delta+min))/delta;	/* 0 <= t < 1 */
						buf = (1.0-t)*xlat[bin] + t*xlat[bin+1];

						switch (img->imgtype)
						{
							case B_CHAR:
							{
								unsigned char* p = img->data;
								c = (int) (0.5+buf);
								if (c<0) c=0; else if (c>255) c=255;
								p[x + img->xmax*(y + img->ymax*z)]=c;
								break;
							}
							case B_SHORT:
							{
								short* p = img->data;
								int s = (int) (0.5+buf);
								p[x + img->xmax*(y + img->ymax*z)]=s;
								break;
							}
							case B_FLOAT:
							{
								float* p = img->data;
								p[x + img->xmax*(y + img->ymax*z)]=buf;
								break;
							}
							case B_RGB:
							{
								rgb = rgbptr[x + img->xmax*(y + img->ymax*z)];
								bin = (int) ((rgb.red - min)/delta);
								t = (rgb.red - (bin*delta+min))/delta;	/* 0 <= t < 1 */
								c = (int)((1.0-t)*xlat[bin] + t*xlat[bin+1]+0.5);
								if (c<0) c=0; else if (c>255) c=255;
								rgb.red = c;
								bin = (int) ((rgb.green - min)/delta);
								t = (rgb.green - (bin*delta+min))/delta;	/* 0 <= t < 1 */
								c = (int)((1.0-t)*xlat[bin+numbins] + t*xlat[bin+1+numbins]+0.5);
								if (c<0) c=0; else if (c>255) c=255;
								rgb.green = c;
								bin = (int) ((rgb.blue - min)/delta);
								t = (rgb.blue - (bin*delta+min))/delta;	/* 0 <= t < 1 */
								c = (int)((1.0-t)*xlat[bin+2*numbins] + t*xlat[bin+1+2*numbins]+0.5);
								if (c<0) c=0; else if (c>255) c=255;
								rgb.blue = c;								
								rgbptr[x + img->xmax*(y + img->ymax*z)]=rgb;
								break;
							}
						}
					}
				}
			}
			
	make_minmax (img);
	img->uid++;
	
}


/* Local histogram equalization. Each pixel gets individually equalized
   in its n by n neighborhood.
   TODO: Ugly with RGB
   TODO: Only 2D slice-by-slice operation
   TODO: Needs to be made *much* more efficient */


void local_histeq (image_type* img, int numbins, int ngbh, int oorflg)
{
int bx1,by1,bx2,by2,x0,y0,x1,y1;
int i,x,y,z,area,cumul,bin,e;
int bins[numbins+1];
float xlat[numbins+1];
float min, max, delta, t, buf;
float Sx,Sxx, frac;
float complete;
image_type slice;
void *dataptr;

	/* Need one slice temp storage */
	
	master_abort=0;
	if (allocate_image (&slice,img->imgtype, img->xmax, img->ymax,1)==-1)
		fatal_memerr ();


	/* If we have a ROI, observe it */
	
	roi_bbox (img, &bx1,&by1,&bx2,&by2);

	imgminmax (img,&min,&max);
	delta = (max-min)/numbins;	/* The width of one bin, starting with min */
	e = DATALEN(img->imgtype);

	for (z=0; z<img->zmax; z++)
	{

		/* initialize the temp slice by copying the original (saves time) */
		
		dataptr = img->data + e*img->xmax*img->ymax*z;
		memcpy (slice.data, dataptr, e*img->xmax*img->ymax);

		for (y=by1; y<=by2; y++)
		{
			if (!macrorun)
			{
				frac = (float)(z+1)*(y-by1)/(float)(by2-by1)/img->zmax;
				indicate_progress (frac);
			}
			// TODO: How do we revert the process when the abort button is pressed?


			for (x=bx1; x<=bx2; x++)
			{
				if (in_roi (x,y))			/* For each pixel inside the ROI */
				{
					
					/* Generate the n x n histogram */
					
					for (i=0; i<=numbins; i++) { bins[i]=0; xlat[i]=0; }
					area=0; Sx=0; Sxx=0;
					
					for (y0=y-ngbh; y0<=y+ngbh; y0++)
						for (x0=x-ngbh; x0<=x+ngbh; x0++)
						{
					
							/* We can't afford to lose points, so use a mirror boundary condition */

							if (x0<0) x1=abs(x0); 
							else if (x0>=img->xmax) x1=img->xmax-(x0-img->xmax);
							else x1=x0;
							if (y0<0) y1=abs(y0); 
							else if (y0>=img->ymax) y1=img->ymax-(y0-img->ymax);
							else y1=y0;
					
							buf = readbuf_flt (img,x0,y0,z);
							if ((!oorflg) || (fabs(buf)>1e-12))
							{
								bin = (int) ((buf - min)/delta);
								if ((bin>=0) && (bin<=numbins)) 
								{
									bins[bin]++;
									area++;
									Sx += buf; Sxx+= SQR(buf);
								}
							}
						}
						
					/* Compute the cumulative histogram which 
						is also the translation table. 
						cumul/area ->1 when i->numbins */

					cumul=0;
					for (i=0; i<=numbins; i++)
					{
						cumul+=bins[i];
						xlat[i] = min+(max-min)*(float)cumul/(float)area;
					}
						
					/* translate this one pixel */

					buf = readbuf_flt (img,x,y,z);
					bin = (int) ((buf - min)/delta);
					t = (buf - (bin*delta+min))/delta;	/* 0 <= t < 1 */
					buf = (1.0-t)*xlat[bin] + t*xlat[bin+1];

					switch (img->imgtype)
					{
						case B_CHAR:
						{
							unsigned char* p = slice.data;
							int c = (int) (0.5+buf);
							if (c<0) c=0; else if (c>255) c=255;
							p[x + img->xmax*y]=c;
							break;
						}
						case B_RGB:
						{
							rgbtriplet* p = slice.data;
							int c = (int) (0.5+buf);
							if (c<0) c=0; else if (c>255) c=255;
							p[x + img->xmax*y]=(rgbtriplet){c,c,c};
							break;
						}
						case B_SHORT:
						{
							short* p = slice.data;
							int s = (int) (0.5+buf);
							p[x + img->xmax*y]=s;
							break;
						}
						case B_FLOAT:
						{
							float* p = slice.data;
							p[x + img->xmax*y]=buf;
							break;
						}
					}


				}
			}

		}

		/* Copy the new slice into the original image */
		
		memcpy (dataptr, slice.data, e*img->xmax*img->ymax);


	}
			
	make_minmax (img);
	img->uid++;
	freebuf (&slice);
	reset_progress ();
	
}



GtkAdjustment *histeq_numbins_spinbutton_adj;
GtkAdjustment *histeq_nghb_spinbutton_adj;
GtkWidget *histeq_zeroisoor_button;
GtkWidget *histeq_local_button;


void on_histeq_ok_clicked (GtkButton *button, gpointer user_data)
{
int oor, local, numbins, nghb;

	undo_nexus (&mainimg);

	oor = gtk_toggle_button_get_active ((GtkToggleButton*)histeq_zeroisoor_button);
	local = gtk_toggle_button_get_active ((GtkToggleButton*)histeq_local_button);
	numbins = histeq_numbins_spinbutton_adj->value;
	nghb = histeq_nghb_spinbutton_adj->value;

	if (local)
	{
		
		dp (1, "Local histogram equalization (box size %dx%d) - may take a while.\n",2*nghb,2*nghb);
		local_histeq (&mainimg,numbins,nghb,oor);
	}
	else
		histeq (&mainimg,numbins, oor);
		
	make_imlib_buffer_from_main_image (cur_slice);
	redraw (1);
	
	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void on_histeq_cancel_clicked (GtkButton *button, gpointer user_data)
{
	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void on_histeq_localbutton_clicked (GtkButton *button, gpointer user_data)
{
int local = gtk_toggle_button_get_active ((GtkToggleButton*)histeq_local_button);

	dp (2, "Histeq: Local mode is now %d\n",local);


}



void create_histeq_wnd1 ()
{
  GtkWidget *histeq_wnd1;
  GtkWidget *histeq_vbox;
  GtkWidget *histeq_table1;
  GtkWidget *histeq_label1, *histeq_label2, *histeq_label3, *histeq_label4;
  GtkWidget *histeq_hseparator1;
  GtkWidget *histeq_hseparator2;
  GtkWidget *histeq_okbutton;
  GtkWidget *histeq_cancelbutton;
  GtkWidget *histeq_numbins_spinbutton;
  GtkWidget *histeq_nghb_spinbutton;

	
	histeq_wnd1 = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (histeq_wnd1), "histeq_wnd1", histeq_wnd1);
	gtk_window_set_title (GTK_WINDOW (histeq_wnd1), "Equalize Histogram");
	gtk_window_set_modal (GTK_WINDOW (histeq_wnd1), TRUE);
	histeq_vbox = vbox_in_window (histeq_wnd1, "histeq_vbox");

	histeq_table1 = MakeTable (histeq_wnd1,histeq_vbox, "histeq_table1",2,6);

	histeq_label1 = MakeLabel (histeq_wnd1,histeq_table1,"Number of bins","histeq_label1",0,0);
	histeq_label2 = MakeLabel (histeq_wnd1,histeq_table1,"Neighborhood radius\n(local only)","histeq_label2",0,1);
	histeq_label3 = MakeLabel (histeq_wnd1,histeq_table1,"Zero is OOR","histeq_label2",0,2);
	histeq_label4 = MakeLabel (histeq_wnd1,histeq_table1,"Local operation","histeq_label3",0,3);

	histeq_numbins_spinbutton_adj = GTK_ADJUSTMENT(gtk_adjustment_new (256, 1, 1024, 1, 10, 0));
	histeq_numbins_spinbutton = MakeSpinbutton (GTK_ADJUSTMENT(histeq_numbins_spinbutton_adj), histeq_wnd1, 
  							histeq_table1, "histeq_numbins_spinbutton", 1,0);
	histeq_nghb_spinbutton_adj = GTK_ADJUSTMENT(gtk_adjustment_new (40, 1, 1024, 1, 10, 0));
	histeq_nghb_spinbutton = MakeSpinbutton (GTK_ADJUSTMENT(histeq_nghb_spinbutton_adj), histeq_wnd1, 
  							histeq_table1, "histeq_ngh_spinbutton", 1,1);
	histeq_zeroisoor_button = Checkbutton_in_table (histeq_wnd1,histeq_table1," ","histeq_oor_button",1,2,NULL);
	histeq_local_button = Checkbutton_in_table (histeq_wnd1,histeq_table1," ","histeq_loc_button",1,3,on_histeq_localbutton_clicked);
 
	histeq_hseparator1 = MakeHsep (histeq_wnd1,histeq_table1,"histeq_hseparator1",0,4);
	histeq_hseparator2 = MakeHsep (histeq_wnd1,histeq_table1,"histeq_hseparator2",1,4);
	
	histeq_okbutton = Button_in_table (histeq_wnd1,histeq_table1,"OK","histeq_okbutton",0,5,on_histeq_ok_clicked);
	histeq_cancelbutton = Button_in_table (histeq_wnd1,histeq_table1,"Cancel","histeq_cancelbutton",1,5,on_histeq_cancel_clicked);

	gtk_signal_connect (GTK_OBJECT (histeq_wnd1), "destroy", GTK_SIGNAL_FUNC (on_histeq_cancel_clicked), histeq_wnd1);


	gtk_widget_show (histeq_wnd1);
  
}


void histeq_main ()
{
	create_histeq_wnd1 ();
}




/****************************************************

	Data conversion. Create a different image type
	(RGB,Char,Short,Float) from the original type
	
*****************************************************/


void dataconv_execute (image_type* img, TYPE_SELECTOR targettype, char dataconv_expand)
{
long x,y,z;
double buf,min,max;
unsigned char e_char;
short e_short;
float e_flt;
rgbtriplet e_rgb;
image_type targ;

	if (img->imgtype == targettype) return;	/* nothing to do if types match */

	if (allocate_image (&targ,targettype,img->xmax,img->ymax,img->zmax)==-1)
		fatal_memerr();
		
	/* Go here if "expand range" was selected. Override as follows:
		- If the target is "float", we NEVER expand (first priority)
		- if the source is RGB, we ALWAYS expand (second priority)
	*/
	
	if (((dataconv_expand) || (img->imgtype==B_RGB)) && (targettype!=B_FLOAT))
	{
		if (img->imgtype==B_FLOAT)
			{	min = img->minval.fmin; max = img->maxval.fmax; }
		else
			{	min = img->minval.imin; max = img->maxval.imax; }

		for (z=0; z<img->zmax; z++)
			for (y=0; y<img->ymax; y++)
				for (x=0; x<img->xmax; x++)
		{
			/* for simplicity, we take the path via "float" since we need to expand anyway */
	
			buf = readbuf_flt (img,x,y,z);
			buf = (buf-min)/(max-min);	/* Scale to the range 0...1 */
			if (targettype==B_SHORT) buf *= 32767.0;
			else if ((targettype==B_RGB) || (targettype==B_CHAR)) buf*=255.0;
			
			/* Now stash it away */
			
			switch (targettype)		/* remember, we excluded float */
			{
				case B_CHAR:
					e_char = (int) (0.5+buf);
					writebuf (&targ, &e_char, x,y,z);
					break;
				case B_SHORT:
					e_short = (int) (0.5+buf);
					writebuf (&targ, &e_short, x,y,z);
					break;
				case B_RGB:
					e_rgb.red = e_rgb.green = e_rgb.blue = (int) (0.5+buf);
					writebuf (&targ, &e_rgb, x,y,z);
					break;
			}
	
		}
	}
	else
	{
		for (z=0; z<img->zmax; z++)
			for (y=0; y<img->ymax; y++)
				for (x=0; x<img->xmax; x++)
		{
			/* Retrieve the source element and convert, 4x4 matrix.
				This may need thresholding.
			   Probably not the most elegant way, but we don't need efficiency here */
			
			if (img->imgtype==B_CHAR)
			{
				readbuf (img, &e_char, x,y,z);
				if (targettype==B_RGB) { e_rgb.red=e_rgb.green=e_rgb.blue=e_char; }
				else if (targettype==B_SHORT) e_short = e_char;
				else if (targettype==B_FLOAT) e_flt = (float)e_char;
			}
			else if (img->imgtype==B_SHORT)	
			{
				readbuf (img, &e_short, x,y,z);
				if ((targettype==B_RGB) || (targettype==B_CHAR))
				{
					if (e_short<0) e_short=0; else if (e_short>255) e_short=255;
				}
				if (targettype==B_RGB) { e_rgb.red=e_rgb.green=e_rgb.blue=e_short; }
				else if (targettype==B_CHAR) e_char = e_short;
				else if (targettype==B_FLOAT) e_flt = (float)e_short;
			}
			else if (img->imgtype==B_FLOAT)	
			{
				readbuf (img, &e_flt, x,y,z);
				if ((targettype==B_RGB) || (targettype==B_CHAR))
				{
					if (e_flt<0.0) e_flt=0; else if (e_flt>255.0) e_flt=255.0;
				}
				else if (targettype==B_SHORT)
				{
					if (e_flt<-32767.0) e_flt=-32767.0; else if (e_flt>32767.0) e_flt=32767.0;
				}
				if (targettype==B_RGB) { e_rgb.red=e_rgb.green=e_rgb.blue=IROUND(e_flt); }
				else if (targettype==B_CHAR) e_char = IROUND(e_flt);
				else if (targettype==B_SHORT) e_short = (int)e_flt;
			}
			else if (img->imgtype==B_RGB)	/* Here, we take the path via float again */
			{
				buf = readbuf_flt (img, x,y,z);	/* No thresholding necessary */
				if (targettype==B_CHAR) e_char = IROUND(buf);
				else if (targettype==B_SHORT) e_short = IROUND(buf);
				else if (targettype==B_FLOAT) e_flt = buf;
			}
			
			/* Now stash it away */
			
			switch (targettype)
			{
				case B_CHAR:
					writebuf (&targ, &e_char, x,y,z);
					break;
				case B_SHORT:
					writebuf (&targ, &e_short, x,y,z);
					break;
				case B_FLOAT:
					writebuf (&targ, &e_flt, x,y,z);
					break;
				case B_RGB:
					writebuf (&targ, &e_rgb, x,y,z);
					break;
			}
	
		}	/* end triple for loop */

	}	/* end selection between expand and keep */
	
	/* Finally, free the old image buffer and hand back the new one */
	
	handback (img, &targ);

}


/* This function handles the special case where a RGB image is 
   expanded into a stack of B_CHAR data with 3 slices.
   This operation changes memory organization, but not memory size */

void dataconv_expand_rgb (image_type* img)
{
unsigned char *p1, *p2;
int z;
long i,idx,xy;

	g_assert (img->imgtype == B_RGB);

	p1 = img->data;	/* source pointer */
	p2 = (unsigned char*)calloc (3*img->xmax*img->ymax, sizeof (unsigned char));	/* One-slice intermediate */
	if (!p2) fatal_memerr();
	
	xy = img->xmax * img->ymax;
	
	for (z=0; z<img->zmax; z++) /* slice by slice */
	{
	
		idx = z * xy;
		for (i=0; i < xy; i++)
		{
			p2[i]       = p1[3*(i+idx)];
			p2[i+xy]    = p1[3*(i+idx)+1];
			p2[i+xy+xy] = p1[3*(i+idx)+2];
		}
		for (i=0; i<3*xy; i++) p1[i+3*idx] = p2[i];
	}
	img->zmax *= 3;
	img->imgtype = B_CHAR;
	if (!macrorun) 
	{
		gtk_adjustment_init_val ((GtkAdjustment*)slice_spinbutton_adj, 0, 0, img->zmax-1);
		cur_slice=0;	
	}
	img->uid = next_uid++;

}



/* This function handles the special case of a 3k-slice uchar image
   converted into a stack of B_RGB data with k slices.
   This operation changes memory organization, but not memory size */

void dataconv_stacktorgb (image_type* img)
{
unsigned char *p1, *p2;
int z;
long i,idx,xy;

	g_assert (img->imgtype == B_CHAR);
	g_assert ( (img->zmax %3)==0);

	p1 = img->data;	/* source pointer with at least 3 slices */
	p2 = (unsigned char*)calloc (3*img->xmax*img->ymax, sizeof (unsigned char));	/* Pretend it's three UCHARs */
	if (!p2) fatal_memerr();
	
	xy = img->xmax * img->ymax;
	
	for (z=0; z<img->zmax; z+=3) /* process groups of three slices */
	{
	
		idx = z * xy;
		for (i=0; i < xy; i++)
		{
			p2[3*i] 	= p1[i+idx];			/* R component from z */
			p2[3*i+1] 	= p1[i+idx+xy];			/* G component from z+1 */
			p2[3*i+2]	= p1[i+idx+xy+xy];		/* B component from z+2 */
		}
		for (i=0; i<3*xy; i++) p1[i+idx] = p2[i];		/* once a slice is done, it can be overwritten */
	}
	img->zmax /= 3;
	img->imgtype = B_RGB;
	if (!macrorun) 
	{
		gtk_adjustment_init_val ((GtkAdjustment*)slice_spinbutton_adj, 0, 0, img->zmax-1);
		cur_slice=0;
	}
	img->uid = next_uid++;

}


/* This function handles RGB -> float via Karhunen-Loeve transform */

void dataconv_karhunen_loeve (image_type* img)
{

	g_assert (img->imgtype == B_RGB);
	karhunen_loeve_transform (img, 0);
}


/* This function transforms a RGB image into a HSV image.
   This is a pixel-by-pixel transform, so keep memory */

void dataconv_hsvconv (image_type* img)
{
int idx;
rgbtriplet* prgb;
rgbtriplet rgb;
hsvtriplet hsv;

	g_assert (img->imgtype==B_RGB);
	prgb = img->data;

	for (idx=0; idx<mainimg.xmax * mainimg.ymax * mainimg.zmax; idx++)
	{
		hsv = rgb_to_hsv (prgb[idx]);
		rgb.red = IROUND((hsv.hue*255.0)/360.0);
		rgb.green=hsv.sat;
		rgb.blue = hsv.val;
		prgb[idx]=rgb;
	}
}


/* This function transforms a HSV image into a RGB image.
   This is a pixel-by-pixel transform, so keep memory */

void dataconv_hsv2rgbconv (image_type* img)
{
int idx;
rgbtriplet* prgb;
rgbtriplet rgb;
hsvtriplet hsv;

	g_assert (img->imgtype==B_RGB);
	prgb = img->data;

	for (idx=0; idx<mainimg.xmax * mainimg.ymax * mainimg.zmax; idx++)
	{
		rgb = prgb[idx];		/* The image is really a HSV image */
		hsv.hue = IROUND((rgb.red*360.0)/255.0);
		hsv.sat = rgb.green;
		hsv.val = rgb.blue;
		rgb = hsv_to_rgb (hsv);
		prgb[idx]=rgb;
	}
}


/* This function transforms a RGB image into a YCrCb image.
   This is a pixel-by-pixel transform, so keep memory */

void dataconv_ycrcbconv (image_type* img)
{
int idx;
rgbtriplet* prgb;
rgbtriplet ycrcb;

	g_assert (img->imgtype==B_RGB);
	prgb = img->data;

	for (idx=0; idx<mainimg.xmax * mainimg.ymax * mainimg.zmax; idx++)
	{
		ycrcb = rgb_to_ycrcb (prgb[idx]);
		prgb[idx]=ycrcb;
	}
}



/******************************************************************/

image_type* dataconv_source;
GtkWidget *dataconv_label3;
char dataconv_expand;
char dataconv_targettype;




void on_dataconv_cancel_clicked ( GtkButton * button, gpointer data )
{
	gtk_widget_destroy ( GTK_WIDGET ( data ) );
}


void on_dataconv_ok_clicked ( GtkButton * button, gpointer data )
{

	undo_nexus (&mainimg);

	if (dataconv_targettype==4)
		dataconv_stacktorgb (dataconv_source);
	else if (dataconv_targettype==5)
		dataconv_expand_rgb (dataconv_source);
	else if (dataconv_targettype==6)
		dataconv_karhunen_loeve (dataconv_source);
	else if (dataconv_targettype==7)
		dataconv_hsvconv (dataconv_source);
	else if (dataconv_targettype==8)
		dataconv_hsv2rgbconv (dataconv_source);
	else if (dataconv_targettype==9)
		dataconv_ycrcbconv (dataconv_source);
	else if (dataconv_source->imgtype != dataconv_targettype) 
		dataconv_execute (dataconv_source, dataconv_targettype, dataconv_expand);
	
	if (dataconv_source == &mainimg) 
	{
		make_imlib_buffer_from_main_image (cur_slice);
		redraw (0);
	}

	gtk_widget_destroy ( GTK_WIDGET ( data ) );
}

void on_dataconv_rangebtn_clicked ( GtkButton * button, gpointer data )
{

	dataconv_expand = !dataconv_expand;
	if (dataconv_expand) 
		gtk_label_set_text (GTK_LABEL(dataconv_label3), "Dynamically expand");
	else
		gtk_label_set_text (GTK_LABEL(dataconv_label3), "Keep values");

}


void on_dataconv_optionmenu (GtkWidget *widget, gpointer user_data)
{

	dataconv_targettype = *(int*) user_data;
	
}


void create_dataconv_wnd1 (image_type* img)
{
  GtkWidget *dataconv_wnd1;
  GtkWidget *dataconv_vbox;
  GtkWidget *dataconv_table1;
  GtkWidget *dataconv_label1;
  GtkWidget *dataconv_label2;
  GtkWidget *dataconv_label4;
  GtkWidget *dataconv_hseparator1;
  GtkWidget *dataconv_hseparator2;
  GtkWidget *dataconv_okbutton;
  GtkWidget *dataconv_cancelbutton;
  GtkWidget *dataconv_typemenu;
  GtkWidget *dataconv_typemenu_menu;
  GtkWidget *glade_menuitem;
  GtkWidget *dataconv_rangebtn;

  char* typetext[10] = {	"8-bit grayscale (uchar)",			/* Always available */
 						"16-bit grayscale (short)",			/* Always available */
						"24-bit RGB (uchar)",				/* Always available */
						"IEEE float",						/* Always available */
						"Stack to RGB",						/* Only for UCHAR images with a stack size divisible by 3 */
						"RGB to stack",						/* Only for B_RGB image */
						"RGB principal component",			/* Only for B_RGB image */
						"RGB to HSV",						/* Only for B_RGB image */
						"HSV to RGB" ,						/* Only for B_RGB image */
						"RGB to YCrCb" };					/* Only for B_RGB image */

	dataconv_source = img;
	dataconv_expand=1;
	dataconv_targettype = img->imgtype;
	
	dataconv_wnd1 = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (dataconv_wnd1), "dataconv_wnd1", dataconv_wnd1);
	gtk_window_set_title (GTK_WINDOW (dataconv_wnd1), "Data type conversion");
	gtk_window_set_modal (GTK_WINDOW (dataconv_wnd1), TRUE);
	dataconv_vbox = vbox_in_window (dataconv_wnd1, "dataconv_vbox");

	dataconv_table1 = MakeTable (dataconv_wnd1,dataconv_vbox, "dataconv_tabl1",2,5);

	dataconv_label1 = MakeLabel (dataconv_wnd1,dataconv_table1,"Source data type","dataconv_label1",0,0);
	dataconv_label2 = MakeLabel (dataconv_wnd1,dataconv_table1,"Target data type","dataconv_label2",0,1);
	dataconv_label3 = MakeLabel (dataconv_wnd1,dataconv_table1,"Dynamically expand","dataconv_label3",0,2);
	dataconv_label4 = MakeLabel (dataconv_wnd1,dataconv_table1,typetext[img->imgtype],"dataconv_label4",1,0);

	dataconv_hseparator1 = MakeHsep (dataconv_wnd1,dataconv_table1,"dataconv_hseparator1",0,3);
	dataconv_hseparator2 = MakeHsep (dataconv_wnd1,dataconv_table1,"dataconv_hseparator2",1,3);

	dataconv_okbutton = Button_in_table (dataconv_wnd1,dataconv_table1,"OK","dataconv_okbutton",0,4,on_dataconv_ok_clicked);
	dataconv_cancelbutton = Button_in_table (dataconv_wnd1,dataconv_table1,"Cancel","dataconv_cancelbutton",1,4,on_dataconv_cancel_clicked);
	dataconv_rangebtn =  Togglebutton_in_table (dataconv_wnd1,dataconv_table1,"Click to toggle","dataconv_rangebtn",1,2,on_dataconv_rangebtn_clicked);
	
	dataconv_typemenu = Make_Optionmenu (dataconv_wnd1,dataconv_table1,"dataconv_typemenu",1,1);
	gtk_container_set_border_width (GTK_CONTAINER (dataconv_typemenu), 3);
	dataconv_typemenu_menu = gtk_menu_new ();

	glade_menuitem = Option_menuitem (dataconv_typemenu_menu,on_dataconv_optionmenu,typetext[0],&menuopt0);
	glade_menuitem = Option_menuitem (dataconv_typemenu_menu,on_dataconv_optionmenu,typetext[1],&menuopt1);
	glade_menuitem = Option_menuitem (dataconv_typemenu_menu,on_dataconv_optionmenu,typetext[2],&menuopt2);
	glade_menuitem = Option_menuitem (dataconv_typemenu_menu,on_dataconv_optionmenu,typetext[3],&menuopt3);

	if ( (img->imgtype == B_CHAR) && ((img->zmax % 3)==0) )
		glade_menuitem = Option_menuitem (dataconv_typemenu_menu,on_dataconv_optionmenu,typetext[4],&menuopt4);

	if (img->imgtype == B_RGB)
	{
		glade_menuitem = Option_menuitem (dataconv_typemenu_menu,on_dataconv_optionmenu,typetext[5],&menuopt5);
		glade_menuitem = Option_menuitem (dataconv_typemenu_menu,on_dataconv_optionmenu,typetext[6],&menuopt6);
		glade_menuitem = Option_menuitem (dataconv_typemenu_menu,on_dataconv_optionmenu,typetext[7],&menuopt7);
		glade_menuitem = Option_menuitem (dataconv_typemenu_menu,on_dataconv_optionmenu,typetext[8],&menuopt8);
		glade_menuitem = Option_menuitem (dataconv_typemenu_menu,on_dataconv_optionmenu,typetext[9],&menuopt9);
	}
		
	gtk_option_menu_set_menu (GTK_OPTION_MENU (dataconv_typemenu), dataconv_typemenu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (dataconv_typemenu), img->imgtype);

	gtk_signal_connect (GTK_OBJECT (dataconv_wnd1), "destroy", GTK_SIGNAL_FUNC (on_dataconv_cancel_clicked), dataconv_wnd1);


	gtk_widget_show (dataconv_wnd1);
  
}


void dataconv_main ()
{

	create_dataconv_wnd1 (&mainimg);

}
